##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A wrapper around pylibmc to make it not raise memcache errors.

One way to use this is to add 'cache-module-name relstorage.pylibmc_wrapper'
to zope.conf and set the 'cache-servers' parameter as well.
"""
import logging
from functools import wraps

import pylibmc  # pylint:disable=import-error
from pylibmc import Error as MemcachedError  # pylint:disable=no-name-in-module,import-error

log = logging.getLogger(__name__)

def _catching(func):
    name = func.__name__

    @wraps(func)
    def wrapper(*args):
        try:
            return func(*args)
        except MemcachedError as e: # pragma: no cover
            log.debug("%s failed: %s", name, e)
            return None
    return wrapper

class Client(object):
    behaviors = {
        "tcp_nodelay": True,
        "ketama": True,
    }

    min_compress_len = 0

    def __init__(self, servers):
        self._client = pylibmc.Client(servers, binary=True)
        self._client.set_behaviors(self.behaviors)
        if pylibmc.support_compression: # pylint:disable=no-member
            self.min_compress_len = 1000

    @_catching
    def delete(self, key):
        return self._client.delete(key)

    @_catching
    def get(self, key):
        return self._client.get(key)

    @_catching
    def get_multi(self, keys):
        return self._client.get_multi(keys)

    @_catching
    def set(self, key, value):
        return self._client.set(
            key, value, min_compress_len=self.min_compress_len)

    @_catching
    def set_multi(self, d):
        return self._client.set_multi(
            d, min_compress_len=self.min_compress_len)

    @_catching
    def add(self, key, value):
        return self._client.add(
            key, value, min_compress_len=self.min_compress_len)

    @_catching
    def flush_all(self):
        return self._client.flush_all()

    @_catching
    def disconnect_all(self):
        return self._client.disconnect_all()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Stats implementations
"""

from __future__ import absolute_import

from ..stats import AbstractStats
from .._util import query_property

class PostgreSQLStats(AbstractStats):

    # Getting the COUNT(*)  of tables can be very expensive
    # due to the need to examine rows to check their MVCC visibility.
    # This method only promises to get the *approximate* size,
    # so we use the tables the optimizer uses to get stats.
    # With the autovacuum daemon running, this shouldn't get *too* far
    # out of date.
    # (https://www.postgresql.org/docs/11/monitoring-stats.html#PG-STAT-ALL-TABLES-VIEW)
    _get_object_count_queries = (
        "SELECT reltuples::bigint FROM pg_class WHERE relname = 'current_object'",
        "SELECT reltuples::bigint FROM pg_class WHERE relname = 'object_state'"
    )

    _update_object_count_queries = (
        # Only on PG11 can you list more than one table.
        'ANALYZE current_object',
        'ANALYZE object_state'
    )

    _update_object_count_query = query_property('_update_object_count')

    def get_db_size(self):
        """Returns the approximate size of the database in bytes"""
        def get_size(_conn, cursor):
            cursor.execute("SELECT pg_database_size(current_database())")
            return cursor.fetchone()[0]
        return self.connmanager.open_and_call(get_size)

    def large_database_change(self):

        def analyze(_conn, cursor):
            # VACUUM cannot be run inside a transaction block;
            # ANALYZE can be. Both update pg_class.reltuples.
            # VACUUM needs a read table lock, meaning it can be blocked by writes
            # and vice-versa; ANALYZE doesn't appear to need locks.
            cursor.execute(self._update_object_count_query)
            # Depending on the number of pages this touched, the estimate
            # can be better or worse.

        self.connmanager.open_and_call(analyze)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""sqlite3 adapter for RelStorage."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

from zope.interface import implementer

from ..adapter import AbstractAdapter
from ..dbiter import HistoryFreeDatabaseIterator
from ..dbiter import HistoryPreservingDatabaseIterator
from ..interfaces import IRelStorageAdapter
from ..packundo import HistoryFreePackUndo
from ..packundo import HistoryPreservingPackUndo

from . import drivers
from .batch import Sqlite3RowBatcher
from .connmanager import Sqlite3ConnectionManager
from .locker import Sqlite3Locker
from .mover import Sqlite3ObjectMover
from .oidallocator import Sqlite3OIDAllocator
from .schema import Sqlite3SchemaInstaller
from .stats import Sqlite3Stats
from .txncontrol import Sqlite3TransactionControl
from .poller import Sqlite3Poller
from .scriptrunner import Sqlite3ScriptRunner


@implementer(IRelStorageAdapter)
class Sqlite3Adapter(AbstractAdapter):
    """
    Adapter for sqlite.
    """
    driver_options = drivers
    WRITING_REQUIRES_EXCLUSIVE_LOCK = True

    def __init__(self, data_dir, pragmas,
                 options=None, oidallocator=None,
                 locker=None,
                 mover=None,
                 connmanager=None):
        self.data_dir = os.path.abspath(data_dir)
        self.pragmas = pragmas
        self.oidallocator = oidallocator
        self.locker = locker
        self.mover = mover
        self.connmanager = connmanager
        super(Sqlite3Adapter, self).__init__(options)

    def _create(self):
        driver = self.driver
        options = self.options
        if self.connmanager is None:
            self.connmanager = Sqlite3ConnectionManager(
                driver,
                path=os.path.join(self.data_dir, 'main.sqlite3'),
                pragmas=self.pragmas,
                options=options
            )

        if self.mover is None:
            self.mover = Sqlite3ObjectMover(
                driver,
                options=options,
            )
        if self.locker is None:
            self.locker = Sqlite3Locker(
                options,
                driver,
                batcher_factory=Sqlite3RowBatcher)

        if not self.oidallocator:
            self.oidallocator = Sqlite3OIDAllocator(
                os.path.join(self.data_dir, 'oids.sqlite3'),
                # No switching during OID allocation. It holds an exclusive
                # lock anyway.
                driver=drivers.Sqlite3Driver()
            )

        self.runner = Sqlite3ScriptRunner()
        self.poller = Sqlite3Poller(
            self.driver,
            keep_history=self.keep_history,
            runner=self.runner,
            revert_when_stale=options.revert_when_stale,
            transactions_may_go_backwards=False
        )

        self.txncontrol = Sqlite3TransactionControl(
            connmanager=self.connmanager,
            poller=self.poller,
            keep_history=self.keep_history,
            Binary=driver.Binary,
        )


        self.schema = Sqlite3SchemaInstaller(
            driver=driver,
            oid_allocator=self.oidallocator,
            connmanager=self.connmanager,
            runner=self.runner,
            keep_history=self.keep_history
        )

        self.stats = Sqlite3Stats(
            self.connmanager,
            self.keep_history
        )

        if self.keep_history:
            self.packundo = HistoryPreservingPackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            self.dbiter = HistoryPreservingDatabaseIterator(
                driver,
            )
        else:
            self.packundo = HistoryFreePackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            # TODO: Subclass for this.
            self.packundo._lock_for_share = 'FOR KEY SHARE OF object_state'
            self.dbiter = HistoryFreeDatabaseIterator(
                driver,
            )

    def new_instance(self):
        inst = type(self)(
            self.data_dir,
            self.pragmas,
            options=self.options,
            oidallocator=self.oidallocator.new_instance(),
            locker=self.locker,
            mover=self.mover,
            connmanager=self.connmanager,
        )
        return inst



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
MVCC tracking for cache objects.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from logging import DEBUG as LDEBUG

from zope.interface import implementer

from relstorage._compat import iterkeys
from relstorage._compat import OID_TID_MAP_TYPE as OidTMap
from relstorage._compat import OidTMap_difference
from relstorage._compat import OidTMap_multiunion
from relstorage._compat import OidTMap_intersection
from relstorage._compat import OID_SET_TYPE as OidSet
from relstorage._compat import iteroiditems
from relstorage._compat import IN_TESTRUNNER
from relstorage._util import log_timed
from relstorage._util import positive_integer
from relstorage._util import TRACE as LTRACE
from relstorage._util import get_time_from_environ
from relstorage._mvcc import DetachableMVCCDatabaseCoordinator
from relstorage.options import Options
from relstorage.interfaces import IMVCCDatabaseViewer

from .interfaces import IStorageCacheMVCCDatabaseCoordinator

logger = __import__('logging').getLogger(__name__)

DEBUG = __debug__ and IN_TESTRUNNER

#: The maximum amount of time, in seconds, we will spend polling the
#: database for OIDS when restoring the cache. If this time is exceeded,
#: we will only partially use the cache. This is not set by default because
#: doing so introduces a slight speed penalty to the polling process.
POLL_TIMEOUT = get_time_from_environ('RS_CACHE_POLL_TIMEOUT', None)

###
# Notes on in-process concurrency:
#
# Where possible, we rely on atomic primitive operations of fundamental types.
# For example, we rely on ``dict[key] = value`` being atomic.
# On CPython we use BTrees which are implemented in C and don't release the GIL
# because they're always resident, so this holds true. On PyPy we use
# native dicts and this also holds true. On CPython if we used PURE_PYTHON
# BTrees, this would *not* hold true, so we also use dicts.
###

# pylint:disable=too-many-lines

class _TransactionRangeObjectIndex(OidTMap):
    """
    Holds the portion of the object index visible to transactions <=
    ``highest_visible_tid``.

    Initialized to be either empty, or with the *complete* index of
    objects <= ``highest_visible_tid`` and greater than
    ``complete_since_tid``. (That is, you cannot pass an 'ignore_txn'
    value to the poller. TODO: Workon that. Maybe there's a way.)

    ``highest_visible_tid`` must always be given, but
    ``complete_since_tid`` may be null.

    These attribute, once set, will never change. We may add data
    prior to and including ``complete_since_tid`` as we access it, but we have no
    guarantee that it is complete.

    When ``complete_since_tid`` and ``highest_visible_tid`` are the same
    """
    __slots__ = (
        'highest_visible_tid',
        'complete_since_tid',
        'accepts_writes',
    )

    # When the root node of a BTree splits (outgrows ``max_internal_size``),
    # it creates a new BTree object to be its child by calling ``type(self)()``
    # That doesn't work if you have required arguments.

    def __init__(self, highest_visible_tid=0, complete_since_tid=None, data=()):
        assert complete_since_tid is None or highest_visible_tid >= complete_since_tid
        self.highest_visible_tid = highest_visible_tid
        self.complete_since_tid = complete_since_tid
        self.accepts_writes = True

        OidTMap.__init__(self, data)

        if self:
            # Verify the data matches what they told us.
            # If we were constructed with data, we must be complete.
            # Otherwise we get built up bit by bit.
            if DEBUG:
                assert self.complete_since_tid
                self.verify()
        else:
            # If we had no changes, then either we polled for the same tid
            # as we got, or we didn't try to poll for changes at all.
            assert complete_since_tid is None or complete_since_tid == highest_visible_tid, (
                complete_since_tid, highest_visible_tid
            )

    def verify(self, initial=True):
        # Check that our constraints are met
        if not self or not __debug__:
            return

        max_stored_tid = self.max_stored_tid()
        min_stored_tid = self.min_stored_tid()
        hvt = self.highest_visible_tid
        if max_stored_tid > hvt:
            raise TypeError("max_stored_tid should be <= hvt")
        if min_stored_tid < 0:
            raise TypeError("Underflow error: min TID must be >= 0")

        if initial:
            # This is only true at startup. Over time we can add older entries.
            assert self.complete_since_tid is None or min_stored_tid > self.complete_since_tid, (
                min_stored_tid, self.complete_since_tid)

    def complete_to(self, newer_bucket):
        """
        Given an incomplete bucket (this object) and a possibly-complete bucket for the
        same or a later TID, merge this one to hold the same data and be complete
        for the same transaction range.

        This bucket will be complete for the given bucket's completion, *if* the
        given bucket actually had a different tid than this one. If the given
        bucket was the same tid, then nothing changed and we can't presume
        to be complete.
        """
        assert not self.complete_since_tid
        assert newer_bucket.highest_visible_tid >= self.highest_visible_tid
        self.update(newer_bucket)
        if newer_bucket.highest_visible_tid > self.highest_visible_tid:
            self.highest_visible_tid = newer_bucket.highest_visible_tid
            self.complete_since_tid = newer_bucket.complete_since_tid

    def merge_same_tid(self, bucket):
        """
        Given an incoming complete bucket for the same highest tid as this bucket,
        merge the two into this object.
        """
        assert bucket.highest_visible_tid == self.highest_visible_tid
        self.update(bucket)
        if bucket.complete_since_tid < self.complete_since_tid:
            self.complete_since_tid = bucket.complete_since_tid

    def merge_older_tid(self, bucket):
        """
        Given an incoming complete bucket for an older tid than this bucket,
        merge the two into this object.

        Because we're newer, entries in this bucket supercede objects
        in the incoming data.

        If the *bucket* was complete to a transaction earlier than the transaction
        we're complete to, we become complete to that earlier transaction.

        Does not modify the *bucket*.
        """
        assert bucket.highest_visible_tid <= self.highest_visible_tid
        #debug('Diff between self', dict(self), "and", dict(bucket), items_not_in_self)
        # bring missing data into ourself, being careful not to overwrite
        # things we do have.
        self.update(bucket.items_not_in(self))
        if bucket.complete_since_tid and bucket.complete_since_tid < self.complete_since_tid:
            self.complete_since_tid = bucket.complete_since_tid

    # These raise ValueError if the map is empty
    if not hasattr(OidTMap, 'maxKey'):
        maxKey = lambda self: max(iterkeys(self))

    if hasattr(OidTMap, 'itervalues'): # BTrees, or Python 2 dict
        def maxValue(self):
            return max(self.itervalues())
        def minValue(self):
            return min(self.itervalues())
    else:
        def maxValue(self):
            return max(self.values())
        def minValue(self):
            return min(self.values())

    if not hasattr(OidTMap, 'iteritems'):
        iteritems = OidTMap.items

    def items_not_in(self, other):
        """
        Return the ``(k, v)`` pairs in self whose ``k`` is not found in *other*
        """
        return OidTMap_difference(self, other)

    max_stored_tid = maxValue
    min_stored_tid = minValue

    def __repr__(self):
        return '<%s at 0x%x hvt=%s complete_after=%s len=%s readonly=%s>' % (
            self.__class__.__name__,
            id(self),
            self.highest_visible_tid,
            self.complete_since_tid,
            len(self),
            not self.accepts_writes,
        )

@implementer(IMVCCDatabaseViewer)
class _ObjectIndex(object):
    """
    Collectively holds all the object index visible to transactions <=
    ``highest_visible_tid``.

    The entries in ``maps`` attribute are instances of
    :class:`_TransactionRangeObjectIndex`, and our
    ``highest_visible_tid`` is exactly that of the frontmost map. For
    convenience, we also define this to be our
    ``maximum_highest_visible_tid``. The ``highest_visible_tid`` of
    the final map defines our ``minimum_highest_visible_tid``.

    The following describes the general evolution of this index and
    its MVCC strategy. Some of this is implemented here, but much of
    it is implemented in the :class:`MVCCDatabaseCoordinator`.

    .. rubric:: Initial State

    In the beginning, the frontmost map will be empty and incomplete;
    it will have a ``highest_visible_tid`` but no
    ``complete_since_tid``. (Alternately, if we restored from cache,
    we may have data, but we are definitely still incomplete.) We may
    load new object indices into this map, but only if they're less
    than what our ``highest_visible_tid`` is (this is just a special
    version of the general rule for updates laid out below).

    When a new poll takes place and produces a new change list, in the
    form of a :class:`_TransactionRangeObjectIndex`, it is merged with
    the previous map and replaced into this object. This map does have
    a ``complete_since_tid.``

    After this first map is added, moving forward there should be no
    gaps in ``complete_since_tid`` coverage. There may be overlaps,
    but there must be no gaps. Thus, we have complete coverage from
    ``highest_visible_tid`` in the first map all the way back to
    ``complete_since_tid`` in the second-to-last map.

    If the front of this instance already has data matching that same
    ``highest_visible_tid``, the two maps are merged together
    preserving the lowest ``complete_since_tid`` value. In this case,
    a new instance is *not* returned; the same instance is returned so
    that this map can be shared.

    If the front of this instance has a ``highest_visible_tid`` *less*
    than the new polled tid (that is, there's a new TID in the
    database) all data in the new map with a TID less than the current
    ``highest_visible_tid`` could be discarded (but since we walk the
    maps from front to back to find entries, the more data up front
    the better). In fact, we could limit the poll to just that range
    (use ``highest_visible_tid`` as our ``prev_polled_tid``
    parameter), if we're willing to walk through the current map and
    find all objects whose TID is greater than what we would have used
    for ``prev_polled_tid`` (assuming it's >= the current
    ``complete_since_tid`` value) --- but it's probably better to just
    let the database do that.

    On the chance that the value we wish to poll from
    (``prev_polled_tid``) is <= our current ``complete_since_tid``,
    the new map can completely replace the current front of the map
    (merging in any extra data).

    Instead of updating the frontmost map when new a new entry is
    presented, we update the oldest map that can see the new entry.
    The older (farther to the right, or towards the end of the
    ``maps`` list), the more likely it is to be shared by the most
    transactions. (That does have the unfortunate effect of making it
    take longer to get a hit since we have to probe further back in
    the map; maybe we want to update both frontmost and backmost?)

    When a transaction ends and polls, such that the
    ``MVCCDatabaseCoordinator.minimum_highest_visible_tid`` increments
    (which also corresponds to the last reference to that
    ``TransactionRangeObjectIndex`` going away), that map is removed
    from the list (where it was the end), and any unique data in it is
    added to the map at the new back of the list. (The object that was
    removed is *cleared* and marked as *inactive*: other transactions
    will still have that same object on their list, but they shouldn't
    use it anymore to store new index entries.)

    This last map will keep growing. We shrink it in one way:

        - When we merge the back two maps, we examine the map being
          removed. This is a snapshot of the database no longer in
          use. We first remove any entries that have newer data in a
          more recent snapshot. Then, any remaining unique entries
          (which by definition are the current state of the affected
          objects) are simply removed (note that we don't search for
          duplicates). Crucially, to obtain decent hit rates, those
          dropped index entries are re-cached with a special key
          indicating that they haven't changed in a long time;
          operations that update the cache with new data must
          invalidate this special key. If an object is not found in
          this index, it can then be looked for under this special
          key. Even though the index data is not complete, this
          "frozen" object can still be found.

    When merging this last map, we have an excellent time to purge
    entries from the cache: Any OIDs stored in the last map (the one
    being removed) who have entries in other maps has their stored
    OID/TID from the last map removed. These have changed, and by
    definition their old state is not visible to any other connection.

    TODO: The merging and recaching might be expensive? Perhaps we
    want to defer it until some threshold is reached?
    """

    __slots__ = (
        'maps',
    )

    def __init__(self, highest_visible_tid, complete_since_tid=None, data=()):
        """
        An instance is created with the first poll, giving us our
        initial TID. It may optionally have data retrieved from
        previously saving the map.
        """
        initial_bucket = _TransactionRangeObjectIndex(highest_visible_tid, None, ())
        initial_bucket.update(data)
        initial_bucket.complete_since_tid = complete_since_tid
        initial_bucket.verify(initial=False)
        # Maps are read from 0...N, so newest bucket must be first.
        self.maps = [initial_bucket]

    def __repr__(self):
        return '<%s at 0x%x maxhvt=%s minhvt=%s cst=%s depth=%s>' % (
            self.__class__.__name__,
            id(self),
            self.maximum_highest_visible_tid,
            self.minimum_highest_visible_tid,
            self.complete_since_tid,
            self.depth,
        )

    def stats(self):
        return {
            'depth': self.depth,
            'hvt': self.maximum_highest_visible_tid,
            'min hvt': self.minimum_highest_visible_tid,
            'total OIDS': self.total_size,
            'unique OIDs': len(self.keys()),
        }

    def keys(self):
        return OidTMap_multiunion(self.maps)

    def __getitem__(self, oid):
        for mapping in self.maps:
            try:
                # TODO: Microbenchmark this. Which is faster,
                # try/catch or .get()?
                return mapping[oid]
            except KeyError:
                pass
        # No data. Could it be frozen? We'll let the caller decide.
        return None

    def __contains__(self, oid):
        return any(oid in m for m in self.maps)

    def __setitem__(self, oid, tid):
        # Silently discard things that are too new.
        # Something like loadSerial for conflict resolution might do this;
        # we'll see the object from the future when we poll for changes.

        # Because each successive map is supposed to be a delta, we only
        # store this in the first possible map.

        # TODO: Maybe this should always store into the master object?

        for mapping in reversed(self.maps):
            mapping_hvt = mapping.highest_visible_tid
            accepts_writes = mapping.accepts_writes
            if not accepts_writes:
                # Closed, but still in our list because this is a shared object.
                continue
            if tid > mapping_hvt:
                # Not visible; everything still to come is newer yet,
                # so we're done.
                break
            if mapping.complete_since_tid and tid > mapping.complete_since_tid:
                assert oid in mapping, (oid, tid, mapping)
                assert mapping[oid] == tid, mapping
                continue
            assert mapping.complete_since_tid is None or tid <= mapping.complete_since_tid
            mapping[oid] = tid
            break


    @property
    def total_size(self):
        """
        The total size of this object is the combined length of all maps;
        this is not the same thing as the length of unique keys.
        """
        return sum(len(m) for m in self.maps)

    @property
    def depth(self):
        return len(self.maps)

    @property
    def highest_visible_tid(self):
        return self.maps[0].highest_visible_tid

    # The maximum_highest_visible_tid of this object, and indeed, of any
    # given map, must never change. It must always match what's
    # visible to the clients it has been handed to.
    maximum_highest_visible_tid = highest_visible_tid

    @property
    def minimum_highest_visible_tid(self):
        return self.maps[-1].highest_visible_tid

    @property
    def complete_since_tid(self):
        # We are complete since the oldest map we have
        # that thinks *it* is complete. This may not necessarily be the
        # last map (but it should be the second to last map!)
        for mapping in reversed(self.maps):
            cst = mapping.complete_since_tid
            if cst is not None:
                return cst

    def verify(self):
        # Each component has values in range.
        for ix in self.maps:
            ix.verify(initial=False)
        # No gaps in completion
        map_iter = iter(self.maps)
        newest_map = next(map_iter)
        for mapping in map_iter:
            __traceback_info__ = newest_map, mapping
            assert newest_map.complete_since_tid <= mapping.highest_visible_tid
            newest_map = mapping

    def with_polled_changes(self, highest_visible_tid, complete_since_tid, changes):

        # Never call this when the poller has specifically said
        # that there are no changes; either the very first poll, or
        # we went backwards due to reverting to a stale state. That's
        # handled at a higher layer.
        assert changes is not None
        assert highest_visible_tid >= self.highest_visible_tid
        assert complete_since_tid is not None

        # First, create the transaction map.
        assert highest_visible_tid and complete_since_tid
        incoming_bucket = _TransactionRangeObjectIndex(highest_visible_tid,
                                                       complete_since_tid,
                                                       changes)
        newest_bucket = self.maps[0]
        oldest_bucket = self.maps[-1]

        # Was this our first poll?
        if newest_bucket is oldest_bucket and oldest_bucket.complete_since_tid is None:
            # Merge the two together and replace ourself.
            assert highest_visible_tid >= oldest_bucket.highest_visible_tid
            if highest_visible_tid == oldest_bucket.highest_visible_tid:
                # Overwrite the incomplete data with the new data, which *may be* complete
                # back to a point. The make the incomplete data become the complete data
                oldest_bucket.complete_to(incoming_bucket)
                #debug("Oldest bucket is now", oldest_bucket, dict(oldest_bucket))
                oldest_bucket.verify()
                return self
            # We need to move forward. Therefore we need a new index.
            # Copy forward any old data we've got and maintain our completion status.
            # But we don't want to lose our connection to the older
            # bucket, because that represents the oldest thing we have indexed.
            # Why merge? That doesn't make much sense. We're no longer a delta.
            # incoming_bucket.merge_older_tid(oldest_bucket)
            other = _ObjectIndex.__new__(_ObjectIndex)
            other.maps = [incoming_bucket, oldest_bucket]
            other.verify()
            return other

        # Special cases:
        #
        # - len(new_bucket) == 0: No changes. Therefore, if this was
        #   not our first poll, it must have returned the exact same
        #   highest_visible_tid as our current highest visible tid. We
        #   just need to set our frontmost map's
        #   ``complete_since_tid`` to the lowest of the two values.
        #   This is just a degenerate case of polling to a matching highest_visible_tid.

        # We can't get a higher transaction id unless something changed.
        # put the other way, if nothing changed, it must be for our existing tid.
        __traceback_info__ = incoming_bucket, newest_bucket
        assert incoming_bucket or (
            not incoming_bucket and highest_visible_tid == newest_bucket.highest_visible_tid
        )

        if highest_visible_tid == newest_bucket.highest_visible_tid:
            #debug("Merging for same tid", newest_bucket, "incoming", incoming_bucket)
            newest_bucket.merge_same_tid(incoming_bucket)
            return self

        # all that's left is to put the new bucket on front of a new object.
        other = _ObjectIndex.__new__(_ObjectIndex)
        other.maps = [incoming_bucket]
        other.maps.extend(self.maps)
        if DEBUG:
            other.verify()
        return other


class _AlreadyClosedLock(object):

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        "Does nothing"

@implementer(IStorageCacheMVCCDatabaseCoordinator)
class MVCCDatabaseCoordinator(DetachableMVCCDatabaseCoordinator):
    """
    Keeps track of the most recent polling data so that
    instances don't make unnecessary polls.

    Shared between all instances of a StorageCache
    in a tree, from the master down.
    """

    # Essentially how many transactions any one viewer is allowed to
    # get behind the leading edge before we cut it off. The next time it polls,
    # it will drop its ZODB object cache. This probably indicates an idle connection
    # without a connection pool idle timeout configured, or a very long-duration
    # read transaction.
    max_allowed_index_depth = positive_integer(
        os.environ.get('RS_CACHE_MVCC_MAX_DEPTH', '100')
    )

    # The total number of entries in the object index we allow
    # before we start cutting off old viewers. This gets set from
    # Options.cache_delta_size_limit
    max_allowed_index_size = 100000
    object_index = None

    def __init__(self, options=None):
        super(MVCCDatabaseCoordinator, self).__init__()
        # There's a tension between blocking as little as possible
        # and making as few polling queries as possible. Polling is when
        # the GIL is released or gevent switches can occur, and potentially
        # allow other threads/greenlets to do useful work. OTOH the more
        # polling queries we do the more (probably overlapping) data we have to read
        # and process.
        #
        # We content ourselves with only locking enough to keep our constraints
        # consistent and allow for potentially overlapped polls. Since we use the previous
        # global poll as our starting point, they should be small.

        options = options or Options()
        # We used to keep two of these...in every connection.
        # If we have many concurrent commits going on, and/or lots of old idle connections,
        # we can surge above this by a substantial amount; as we continue processing transactions,
        # each new one will drop more old viewers, though, and it will start to be reclaimed.
        # Also, lots of it is shared across the connections.
        self.max_allowed_index_size = options.cache_delta_size_limit * 2
        self.log = logger.log

    def stats(self):
        return {
            'registered_viewers': len(self._registered_viewers),
            'oldest viewer': self.minimum_highest_visible_tid,
            'hvt': self.object_index.maximum_highest_visible_tid if self.object_index else None,
            'index': self.object_index.stats() if self.object_index else None,
        }

    @property
    def complete_since_tid(self):
        return -1 if self.object_index is None else self.object_index.complete_since_tid

    def reset_viewer(self, cache):
        cache.object_index = None
        self.change(cache, None)

    def poll(self, cache, conn, cursor):
        with self._lock:
            cur_ix = self.object_index
            # this can mutate without changing the object identity!
            cur_ix_hvt = cur_ix.highest_visible_tid if cur_ix else None

        return self._poll(cache, conn, cursor, cur_ix, cur_ix_hvt)

    def __set_viewer_state_locked(self, cache, index):
        cache.object_index = index
        # attaches the viewer if it was detached.
        self.change(cache,
                    index.highest_visible_tid
                    if index is not None
                    else None)

    def __initial_poll(self, cache, cursor):
        # Initial poll for the world.
        tid = cache.adapter.poller.get_current_tid(cursor)

        new_index = None
        if tid > 0:
            # tid 0 is empty database, no data.
            new_index = _ObjectIndex(tid)

        with self._lock:
            if new_index is not None and self.object_index is None:
                # If we're still none, we got here first, yay us.
                self.object_index = new_index

            self.__set_viewer_state_locked(cache, self.object_index)
        # Regardless whether we or someone else did or did not
        # get an index, the viewer gets our current state, and also
        # gets told to invalidate everything (implicit return None)

    def _poll(self, cache, conn, cursor,
              current_index,
              current_index_hvt):
        # Note that poll_invalidations can raise StaleConnectionError,
        # or it can return (None, old_tid) where old_tid is less than
        # its third parameter (``prev_polled_tid``)
        if current_index is None:
            return self.__initial_poll(cache, cursor)

        # We have begun keeping an object index. But the cache
        # may not yet. (See comments in _ObjectIndex docstring about
        # possible optimizations for that case, or the case when it can use
        # a different poll range.)

        polling_since = current_index_hvt

        # Do a small poll.
        # NOTE: See comment in __init__ about tensions about locking
        # and overlapping polls.
        change_iter, polled_tid = cache.adapter.poller.poll_invalidations(
            conn, cursor,
            polling_since,
        )

        if polled_tid == 0 or polled_tid < polling_since:
            assert change_iter is None
            # Freshly zapped or empty database (tid==0) or stale replica and
            # asked to revert (as opposed to raising
            # ReadConflictError). Mark not just this one, but all
            # other extent indexes as needing a full rebuild.
            with self._lock:
                self.flush_all()
                self.__set_viewer_state_locked(cache, None)
            return None

        # Ok cool, we got data to move us forward.
        # We must be careful to always consume the iterator, even if we exit early
        # (because it could be a server-side cursor holding connection state).
        # So we do that now.
        change_iter = list(change_iter)
        self.log(
            LTRACE,
            "Polled new tid %s since %s with %s changes",
            polled_tid, polling_since, len(change_iter)
        )

        # The *change_index* is the index we'll provide to the viewer; it
        # *must* be built from the data we polled, or at least, from an
        # index with an HVT equal to what we polled. It cannot go into the future
        # of what we polled.
        change_index = None
        # Set *should_vacuum* to true when we're the ones that moved the
        # state forward.
        should_vacuum = False
        installed_index = None
        with self._lock:
            installed_index = self.object_index
            if installed_index is None:
                # Whoops, somebody flushed things while we were working. Alrighty then.
                self.__set_viewer_state_locked(cache, None)
                return None

            if installed_index.highest_visible_tid < polled_tid:
                # it's older than our poll, we take control;
                # but the history of the index must always move forward,
                # so we build it starting from what's currently installed.
                # There could be some overlap. Since we moved the index forward,
                # we can vacuum.
                change_index = self.object_index = installed_index.with_polled_changes(
                    polled_tid,
                    polling_since,
                    change_iter
                )
                change_iter = ()
                should_vacuum = True
            elif installed_index.highest_visible_tid == polled_tid:
                # Cool, no work for us!
                # we can simply discard our poll data.
                change_iter = ()
                change_index = installed_index
            # Otherwise, the data they have is in the future. Whoops!
            # We must build our own change data.
            del installed_index

        # Build changes without the lock.
        if change_index is None:
            # Note that this is going to create a gap in the
            # master object_index. The master object_index
            # may have transaction ranges like:
            # (26, 31]
            # (31, 37]
            # and we have now created an alternate index with
            # (26, 31]
            # (32, 35]
            # and can thus produce a self.minimum_highest_visible_tid
            # of 35, which is not directly answerable with any
            # transaction range in the index.
            # The vacuum process has to be careful to check that
            # the min HVT is answerable *before* it removes a map.
            change_index = current_index.with_polled_changes(
                polled_tid,
                polling_since,
                change_iter
            )
        change_iter = self._find_changes_for_viewer(cache, change_index)

        # Move our MVCC state forward and vacuum while locked.
        with self._lock:
            if self.object_index is None:
                self.__set_viewer_state_locked(cache, None)
                return None

            self.__set_viewer_state_locked(cache, change_index)
            if should_vacuum:
                # Must be sure to vacuum using the state we installed.
                # If it's been replaced, it will only have moved forward
                # and all the rest of the maps are still shared.
                if self.object_index.highest_visible_tid >= polled_tid:
                    self._vacuum(cache, change_index)

        return change_iter

    @staticmethod
    def _find_changes_for_viewer(viewer, object_index):
        """
        Given a freshly polled *object_index*, and the *viewer* that polled
        for it, build a changes iterator.

        Call this **before** updating the viewer's MVCC state, so that
        we know how far back we need to build the changes.

        Does not need to hold the lock, except that the index cannot be
        vacuumed until this process is complete (since we may need that for
        building changes).
        """
        if viewer.highest_visible_tid is None or viewer.detached:
            # Snarf. Old state, and we probably lost track of changes.
            # Whelp, it needs to invalidate all its cached objects (so
            # we must return None), but it can still use our index and
            # poll state going forward; we don't need to go backwards.
            logger.debug(
                "Invalidating all persistent objects for viewer %r (detached? %s)",
                viewer, viewer.detached
            )
            return None

        # Somewhere in the index is a map with the highest visible tid
        # matching the last time this viewer polled. Everything from there
        # forward is a change
        # Note there could be no changes.
        last_poll_time = viewer.highest_visible_tid
        changes = OidTMap()
        change_dicts = []
        for m in object_index.maps:
            if m.highest_visible_tid == last_poll_time:
                break
            change_dicts.append(m)

        while change_dicts:
            # In reverse order, capturing only the most recent change.
            # TODO: Except for that 'ignore_tid' passed to the viewer's
            # poll method, we could very efficiently do this with
            # OidTMap_multiunion with one call to C.
            changes.update(change_dicts.pop())

        return iteroiditems(changes)

    @log_timed
    def _vacuum(self, cache, object_index):
        """
        Handle object index and cache entries for which we no longer
        have a requirement.

        Named for the ``VACUUM`` command and process in PostgreSQL,
        this notices when we are keeping index data for transactions
        that are no longer needed.

        When this function is called, the cache doing the polling
        should have been updated with the new object index (and thus
        ``highest_visible_tid``), and thus released its claim to its
        former TID.

        Usually, ``object_index`` will be the same as ``self.object_index``
        but that's not required to enable vacuuming of divergent indices
        in concurrent situations.

        If the given ``object_index`` has a ``minimum_highest_visible_tid``
        (i.e., the oldest polled transaction) that is now less than
        the oldest polled transaction needed by any extent cache
        registered to this coordinator, we are free to vacuum that
        oldest state.

        Usually, the next state's ``minimum_highest_visible_tid`` will
        match exactly our new required minimum, but in case of divergent
        indices, there may be a gap. We do not want to vacuum in that case
        because, even though the objects with the partially diverged index
        will still be able to read just fine, we may prematurely remove
        cached object states that they need.
        """
        # This is called for every transaction. It needs to be fast, and mindful
        # of what it logs.
        #
        # MVCC can easily develop "gaps", where one lone reader is at
        # the back and all the other readers are up front somewhere,
        # with that chain of maps in between doing no one any good. We
        # should try to squash those gaps.

        # We partly deal with that by deciding to cut off the oldest viewers
        # (probably idle connections sitting in the pool) if the total depth
        # gets too deep, or the total size gets too large. Do that first.
        if (
                object_index.depth > self.max_allowed_index_depth
                or object_index.total_size > self.max_allowed_index_size
        ):
            self.log(
                LDEBUG,
                "Detaching oldest viewers because depth (%s) > max depth (%s) "
                "or total size (%s) > max size (%s)",
                object_index.depth, self.max_allowed_index_depth,
                object_index.total_size, self.max_allowed_index_size,
            )
            self.detach_viewers_at_minimum()

        required_tid = self.minimum_highest_visible_tid # This won't change during our invocation
        local_client = cache.local_client
        self.log(
            LTRACE,
            "Attempting vacuum from %s up to %s",
            object_index.minimum_highest_visible_tid,
            required_tid,
        )
        oids_tids_to_del = OidTMap()
        while 1:
            if object_index.depth == 1:
                # Nothing left to vacuum
                break
            if object_index.minimum_highest_visible_tid == required_tid:
                # Still need this history.
                break
            next_state = object_index.maps[-2]
            if required_tid and next_state.highest_visible_tid > required_tid:
                # The last state isn't quite obsolete, others are still looking
                # at that range. Don't remove it.
                break

            # all remaining valid viewers have highest_visible_tid > this one
            # So any OIDs that exist in both this bucket and any newer bucket with a newer
            # TID can be purged from the local cache because they've been changed.
            obsolete_bucket = object_index.maps.pop()
            # Immediately also mark it as closed before we start mutating its
            # contents. No more storing to this one!
            obsolete_bucket.accepts_writes = False
            newer_oids = object_index.keys()
            in_both = OidTMap_intersection(newer_oids, obsolete_bucket)

            self.log(
                LTRACE,
                "Examining %d old OIDs to see if they've been replaced",
                len(in_both)
            )

            for oid in in_both:
                old_tid = obsolete_bucket[oid]
                newer_tid = object_index[oid]
                # We intersected, we're sure that they're both not None.
                if newer_tid != old_tid:
                    # Note that even though we're removing data from
                    # this bucket that might be in the range that it
                    # claims to have complete index data for, that's
                    # fine: The end result when we put everything back
                    # together is still going to be complete index
                    # data, because the object changed in the future.
                    # This particular transaction chunk won't be complete, but
                    # it's inaccessible.
                    # This is where we should hook in the 'invalidation' tracing.
                    del obsolete_bucket[oid]
                    oids_tids_to_del[oid] = old_tid # These will just keep going up
                    # If we have a shared memcache, we can't be sure everyone
                    # else is done with this key, so we just leave it alone.

            # Now at this point, the obsolete_bucket contains data that we know is
            # either not present in a future map, or is present with exactly the
            # same value. Therefore, at least until someone changes it again,
            # we can consider this data to be frozen. We'll make available each
            # cached value at a special key. There's nothing
            # useful to have in this last bucket and we can throw it away. Note that
            # we do *not* remove the index entries; they're needed to keep
            # the CST in sync for newer transactions that might still be open.
            if obsolete_bucket:
                self.log(LTRACE, "Vacuum: Freezing %s old OIDs", len(obsolete_bucket))
                local_client.freeze(obsolete_bucket)

        if oids_tids_to_del:
            local_client.delitems(oids_tids_to_del)


    def flush_all(self):
        with self._lock:
            self.object_index = None
            self.detach_all()

    def close(self):
        self.clear()
        with self._lock:
            self.object_index = None

    def save(self, cache, save_args):
        if not self.object_index or not self.object_index.maximum_highest_visible_tid:
            # We have never polled or verified anything, don't
            # try to save what we can't validated.
            logger.debug("No index or HVT; nothing to save %s", self.stats())
            return
        # Vacuum, disposing of uninteresting and duplicate data.
        # We should have no viewers, so we eliminated all except the final map.
        self.detach_all()
        self._vacuum(cache, self.object_index)
        # At this point, we now have processed all the extent invalidations.
        # Note that if there was previously saved data that we invalidated,
        # and have vacuumed away from our index now, we won't know to remove it from
        # the cache file. We do that at load time.

        assert self.object_index.depth == 1, (self.object_index, self._registered_viewers)
        # We give that last map to the local client so it knows to write only
        # known-valid data and to dispose of anything invalid.

        checkpoints = None
        max_hvt = self.object_index.maximum_highest_visible_tid
        checkpoints = (
            max_hvt,
            self.complete_since_tid or max_hvt
        )
        local_client = cache.local_client
        return local_client.save(object_index=self.object_index.maps[0],
                                 checkpoints=checkpoints, **save_args)

    def restore(self, adapter, local_client, timeout=None):
        # This method is not thread safe

        # Note that there may have been an arbitrary amount of data in
        # the file that we didn't get to actually store but that still
        # comes back in the object_index; that's ok. If we pick up some
        # database changes we will start dropping this off pretty quickly.
        # TODO: maybe we want to split the incoming data up by transaction,
        # just like we built it that way in memory.

        checkpoints = local_client.restore()
        highest_visible_tid = checkpoints[0] if checkpoints else None

        if highest_visible_tid:
            # We will thus begin polling at the last poll location
            # stored in the data. All loaded rows are treated as frozen.
            # We won't write them back out.

            self.object_index = _ObjectIndex(highest_visible_tid)
            self.__poll_old_oids_and_remove(adapter, local_client, timeout or POLL_TIMEOUT)

    @log_timed
    def __poll_old_oids_and_remove(self, adapter, local_client, timeout):
        from relstorage.adapters.connmanager import connection_callback
        from relstorage.adapters.interfaces import AggregateOperationTimeoutError

        cached_oids = OidSet(local_client.keys())
        # In local tests, this function executes against PostgreSQL 11 in .78s
        # for 133,002 older OIDs; or, .35s for 57,002 OIDs against MySQL 5.7.
        # In one production environment of 800,000 OIDs with a 98% survival rate,
        # using MySQL 5.7 takes an average of about 11s.
        # However, it has been observed that in some cases, presumably when the database
        # is under intense IO stress, this can take 400s for 500,000 OIDS:
        # since the ``current_object_tids`` batches in groups of 1024, that works out to
        # .75s per SQL query. Not good. Hence the ability to set a timeout.
        logger.info("Polling %d oids stored in cache with SQL timeout %r",
                    len(cached_oids), timeout)

        @connection_callback(isolation_level=adapter.connmanager.isolation_load,
                             read_only=True)
        def poll_cached_oids(_conn, cursor):
            # type: (Any, Any) -> Dict[Int, Int]
            """Return mapping of {oid_int: tid_int}"""
            try:
                return adapter.mover.current_object_tids(cursor, cached_oids,
                                                         timeout=timeout)
            except AggregateOperationTimeoutError as ex:
                # If we time out, we can at least validate the results we have
                # so far.
                logger.info(
                    "Timed out polling the database for %s oids; will use %s partial results",
                    len(cached_oids), len(ex.partial_result)
                )
                return ex.partial_result

        current_tids = adapter.connmanager.open_and_call(poll_cached_oids)
        current_tid = current_tids.get
        polled_invalid_oids = OidSet()
        cache_is_correct = local_client._cache.contains_oid_with_tid

        for oid_int in cached_oids:
            if not cache_is_correct(oid_int, current_tid(oid_int)):
                polled_invalid_oids.add(oid_int)

        logger.info("Polled %d older oids stored in cache (%d found in database); %d survived",
                    len(cached_oids), len(current_tids),
                    len(cached_oids) - len(polled_invalid_oids))
        local_client.remove_invalid_persistent_oids(polled_invalid_oids)



# -*- coding: utf-8 -*-
"""
Elements of select queries.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .query import Query
from .query import ColumnList
from .query import OrderBy
from .query import WhereMixin
from ._util import copy
from ._util import Resolvable

from .ast import TextNode
from .ast import resolved_against
from .expressions import EmptyExpression


class _SelectColumns(ColumnList):

    def __compile_visit__(self, compiler):
        compiler.visit_select_list_csv(self._columns)

    def as_select_list(self):
        return self


class Select(Query,
             Resolvable,
             WhereMixin):
    """
    A Select query.

    When instances of this class are stored in a class dictionary,
    they function as non-data descriptors: The first time they are
    accessed, they *bind* themselves to the instance and select the
    appropriate SQL syntax and compile themselves into a string.
    """

    _distinct = EmptyExpression()

    _order_by = EmptyExpression()
    _limit = None
    _for_update = None
    _nowait = None

    def __init__(self, table, *columns):
        self.table = table
        self.columns = columns
        if columns:
            self.column_list = _SelectColumns(resolved_against(columns, table))
        else:
            self.column_list = table
        self.c = self.column_list

    _bind_vars_ignored = ('column_list',)

    def _bound_to(self, context, dialect):
        super(Select, self)._bound_to(context, dialect)
        if self.columns:
            # Need to re-resolve, it's possible our table changed.
            self.column_list = _SelectColumns(
                resolved_against(self.columns, self.table)
            )
        return self

    def order_by(self, expression, dir=None):
        expression = expression.resolve_against(self.table)
        s = copy(self)
        s._order_by = OrderBy(expression, dir)
        return s

    def unordered(self):
        if not self._order_by:
            return self
        s = copy(self)
        del s._order_by
        return s

    def limit(self, literal):
        s = copy(self)
        s._limit = literal
        return s

    def for_update(self):
        s = copy(self)
        s._for_update = 'FOR UPDATE'
        return s

    def nowait(self):
        s = copy(self)
        s._nowait = 'NOWAIT'
        return s

    def distinct(self):
        s = copy(self)
        s._distinct = TextNode('DISTINCT')
        return s

    def join_kind(self, kind):
        s = copy(self)
        s.table = s.table.join_kind(kind)
        return s

    def is_unconstrained(self):
        "Does this query have nothing after the basic table?"
        return not self._order_by and not self._where

    def __compile_visit__(self, compiler):
        with compiler.visit_limited_select(self, self._limit):
            compiler.emit_keyword('SELECT')
            compiler.visit(self._distinct)
            compiler.visit_select_list(self.column_list)
            compiler.visit_from(self.table)
            compiler.visit_clause(self._where)
            compiler.visit_clause(self._order_by)
            compiler.visit_limit(self._limit)
            if self._for_update:
                compiler.emit_keyword(self._for_update)
            if self._nowait:
                compiler.emit_keyword(self._nowait)


class Selectable(object):
    """
    Mixin for something that can form the root of a selet query.
    """

    def select(self, *args, **kwargs):
        return Select(self, *args, **kwargs)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for storage implementation utilities.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ZODB.POSException import ReadOnlyError

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase

from ..interfaces import IStaleAware

from .. import util

class Foo(object):

    @util.stale_aware
    def method(self, a1, a2=42):
        "Nothing"

    @util.stale_aware
    @util.writable_storage_method
    def writable(self, a1, a2=42):
        return a1, a2


class TestStaleAware(TestCase):

    def test_provides(self):

        method = Foo().method

        assert_that(method, validly_provides(IStaleAware))

    def test_stale(self):

        method = Foo().method

        stale = method.stale(KeyError) # pylint:disable=no-member
        assert_that(stale, validly_provides(IStaleAware))

        with self.assertRaises(KeyError):
            stale()

        stale_again = stale.stale(ValueError)
        self.assertIs(stale_again, stale)

        not_stale = stale.no_longer_stale()
        self.assertIs(not_stale, method)
        not_stale2 = not_stale.no_longer_stale()
        self.assertIs(not_stale2, method)

    def test_writable_under_stale(self):

        class Storage(object):

            read_only = False
            _read_only_error = ReadOnlyError

            def isReadOnly(self):
                return self.read_only

        storage = Storage()
        util.copy_storage_methods(storage, Foo())
        # pylint:disable=no-member
        self.assertEqual(
            (42, 13),
            storage.writable(42, a2=13))

        w = storage.writable.stale(KeyError)
        with self.assertRaises(KeyError):
            w(42)


        storage = Storage()
        storage.read_only = True

        util.copy_storage_methods(storage, Foo())

        with self.assertRaises(ReadOnlyError):
            storage.writable(42, a=36)

        # read-only trumps stale.
        self.assertFalse(hasattr(storage.writable, 'stale'))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Interfaces for top-level RelStorage components.

These interfaces aren't meant to be considered public, they exist to
serve as documentation and for validation of RelStorage internals.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import Interface
from zope.interface import Attribute

import ZODB.interfaces
from ZODB import POSException

# pylint:disable=inherit-non-class, no-self-argument, no-method-argument
# pylint:disable=too-many-ancestors
try:
    from zope.schema import Tuple
    from zope.schema import Object
    from zope.schema import Bool
    from zope.schema import Int
    from zope.interface.common.interfaces import IException
except ImportError: # pragma: no cover
    # We have nti.testing -> zope.schema as a test dependency; but we
    # don't have it as a hard-coded runtime dependency because we
    # don't want to force a version on consumers of RelStorage.
    class _Field(Attribute):
        __allowed_kw__ = ()
        def __init__(self, description, required=False, **kwargs):
            description = "%s (required? %s)" % (description, required)
            for k in self.__allowed_kw__:
                kwargs.pop(k, None)
            if kwargs:
                raise TypeError("Unexpected keyword arguments: %r" % (kwargs,))
            Attribute.__init__(self, description)

    class Tuple(_Field):
        __allowed_kw__ = ('value_type', )

    class Object(_Field):
        def __init__(self, schema, description=''):
            description = "%s (Must provide %s)" % (description, schema)
            super(Object, self).__init__(description)

    Bool = _Field
    Int = _Field

    class Factory(_Field):
        def __init__(self, schema, description='', **kwargs):
            description = "%s (Must implement %s)" % (description, schema)
            super(Factory, self).__init__(description, **kwargs)


    IException = Interface
else:
    from zope.schema.interfaces import SchemaNotProvided as _SchemaNotProvided
    from zope.schema import Field as _Field

    class Factory(_Field):
        def __init__(self, schema, **kw):
            self.schema = schema
            _Field.__init__(self, **kw)

        def _validate(self, value):
            super(Factory, self)._validate(value)
            if not self.schema.implementedBy(value):
                raise _SchemaNotProvided(self.schema, value).with_field_and_value(self, value)

__all__ = [
    'Tuple',
    'Object',
    'Bool',
    'TID',
    'OID',
    'IException',
    'Factory',
    'IMVCCDatabaseCoordinator',
    'IMVCCDatabaseViewer'
]


class OID(Int):
    """
    A ZODB object identifier, represented as a 64-bit integer.
    """

class TID(Int):
    """
    A ZODB transaction identifier, represented as a 64-bit integer.

    Traditionally, ZODB TIDs are created and derived using
    `persistent.timestamp.TimeStamp`, which is a reference to the current
    `time.time` value.
    """

###
# Efficiently handling multiple views of a database
# within a process.
###

class IMVCCDatabaseViewer(Interface):
    """
    A component that has a consistent, point-in-time view of a
    database.

    This is implemented using an RDBMS connection (session) with
    ``REPEATABLE READ`` or higher isolation level.

    In the context of ZODB, this means that this view contains all the
    data for a particular transaction identifier (TID, also "revision"
    or "revid"; the contents of a particular persistent object's
    ``_p_serial``) and the transactions that come before it (lower
    TIDs), but not any newer (higher numbered) transactions that may
    exist.

    The highest available TID is updated between transactions.

    Viewers are associated with a :class:`IMVCCDatabaseCoordinator`.
    """

    highest_visible_tid = TID(
        description=(
            u"""
            The identifier of the most recent transaction viewable to
            this component. A value of ``None`` means that we have no
            idea what transactions even exist.
            """),
        required=False)

class IDetachableMVCCDatabaseViewer(IMVCCDatabaseViewer):
    """
    An MVCC database viewer that can optionally be marked as detached
    from ongoing MVCC operations.

    Viewer may be detached when their coordinator projects it to
    become too expensive to keep them updated (for example, they have
    fallen too far behind the leading edge of the database, possibly
    due to sitting idle). Views are detached instead of destroying
    their ``highest_visible_tid`` because it's possible they may
    actively be working on their view of the database (for example, a
    long running read transaction --- this is exactly what ZODB
    historical connections become).

    Once detached, such a view cannot be updated to a more current
    state. Instead, it must be recreated and all of its cached state
    thrown away.

    Detached viewers are not counted in their coordinator's
    ``minimum_highest_visible_tid``.
    """

    detached = Bool(
        description=u"Is this object detached?"
    )

class IMVCCDatabaseCoordinator(Interface):
    """
    A component that handles tracking multiple `IMVCCDatabaseViewer`
    components fulfilling the same role.

    These components would be created by calling
    :meth:`ZODB.interfaces.IMVCCStorage.new_instance` on the
    `relstorage.interfaces.IRelStorage` owned by the
    `ZODB.interfaces.IDatabase` object. There will be one for each
    `ZODB.interfaces.IConnection` object in a pool.

    By tracking all existing components for a database within the same
    process, we can know what the maximum and minimum visible TIDs are
    within the process. When the minimum visible TID is incremented,
    we have an opportunity to take actions such as freeing data no
    longer needed (because it has been updated in a subsequent
    transaction and we now know the old states aren't visible to any
    current connections.)
    """

    def register(viewer):
        """
        Register the *viewer* to be tracked by this object.

        A matching call to :meth:`unregister` is expected.
        """

    def unregister(viewer):
        """
        Stop tracking the *viewer*.
        """

    maximum_highest_visible_tid = TID(
        description=(
            u"""
             Across all tracked components, report the current highest
             visible tid. This is the most recent transaction that can
             be seen in this process.
             """),
        required=False)

    minimum_highest_visible_tid = TID(
        description=(
            u"""
             Across all tracked components, report the current minimum
             highest visible tid. This is the oldest transaction potentially
             being viewed in this process.
             """),
        required=False)



class IRelStorage(
        ZODB.interfaces.IMVCCAfterCompletionStorage, # IMVCCStorage <- IStorage
        ZODB.interfaces.IMultiCommitStorage,  # mandatory in ZODB5, returns tid from tpc_finish.
        ZODB.interfaces.IStorageRestoreable,  # tpc_begin(tid=) and restore()
        ZODB.interfaces.IStorageIteration,    # iterator()
        ZODB.interfaces.IStorageCurrentRecordIteration, # record_iternext()
        ZODB.interfaces.ReadVerifyingStorage, # checkCurrentSerialInTransaction()
        IMVCCDatabaseViewer,
):
    """
    The relational storage backend.

    These objects are not thread-safe.

    Instances may optionally implement some other interfaces,
    depending on their configuration. These include:

    - :class:`ZODB.interfaces.IBlobStorage` and :class:`ZODB.interfaces.IBlobStorage`
      if a ``blob-dir`` is configured.
    - :class:`ZODB.interfaces.IStorageUndoable` if ``keep-history`` is true.

    """


class POSKeyError(POSException.POSKeyError):
    """
    A POSKeyError that records some extra information
    and prints it.
    """

    extra = None

    def __init__(self, oid_bytes, extra=None, **kwargs):
        if extra is None:
            extra = kwargs
        self.extra = extra
        if extra:
            super(POSKeyError, self).__init__(oid_bytes, extra)
        else:
            super(POSKeyError, self).__init__(oid_bytes)

    def __str__(self):
        s = super(POSKeyError, self).__str__()
        if self.extra:
            s = "%s (%r)" % (s, self.extra)
        return s



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Implementation of `copyTransactionsFrom`.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import tempfile
import time

from ZODB.blob import is_blob_record
from ZODB.utils import u64 as bytes8_to_int64
from ZODB.utils import cp as copy_blob
from ZODB.POSException import POSKeyError


logger = __import__('logging').getLogger(__name__)

class Copy(object):

    __slots__ = (
        'blobhelper',
        'tpc',
        'restore',
    )

    def __init__(self, blobhelper, tpc, restore):
        self.blobhelper = blobhelper
        self.tpc = tpc
        self.restore = restore

    # Time in seconds between progress logging.
    log_interval = 60

    # Number of transactions to copy before checking if we should log.
    log_count = 10

    def copyTransactionsFrom(self, other):
        # pylint:disable=too-many-locals,too-many-statements,too-many-branches
        # adapted from ZODB.blob.BlobStorageMixin
        begin_time = time.time()
        log_at = begin_time + self.log_interval
        txnum = 0
        total_size = 0
        blobhelper = self.blobhelper
        tpc = self.tpc
        restore = self.restore

        logger.info("Counting the transactions to copy.")
        other_it = other.iterator()
        logger.debug("Opened the other iterator: %s", other_it)
        try:
            num_txns = len(other_it)
            if num_txns == 0:
                # Hmm, that shouldn't really be right, should it?
                # Try the other path.
                raise TypeError()
        except TypeError:
            logger.debug("Iterator %s doesn't support len()", other_it)
            num_txns = 0
            for _ in other_it:
                num_txns += 1
            other_it.close()
            other_it = other.iterator()
        logger.info("Copying %d transactions", num_txns)

        tmp_blobs_to_rm = []
        for trans in other_it:
            txnum += 1
            num_txn_records = 0

            tpc.tpc_begin(trans, trans.tid, trans.status)
            for record in trans:
                blobfile = None
                if is_blob_record(record.data):
                    try:
                        blobfile = other.openCommittedBlobFile(
                            record.oid, record.tid)
                    except POSKeyError:
                        pass
                if blobfile is not None:
                    fd, name = tempfile.mkstemp(
                        suffix='.tmp',
                        dir=blobhelper.temporaryDirectory()
                    )
                    tmp_blobs_to_rm.append(name)
                    logger.debug("Copying %s to temporary blob file %s for upload",
                                 blobfile, name)

                    with os.fdopen(fd, 'wb') as target:
                        copy_blob(blobfile, target)
                    blobfile.close()
                    restore.restoreBlob(record.oid, record.tid, record.data,
                                        name, record.data_txn, trans)
                else:
                    restore.restore(record.oid, record.tid, record.data,
                                    '', record.data_txn, trans)
                num_txn_records += 1
                if record.data:
                    total_size += len(record.data)
            tpc.tpc_vote(trans)
            tpc.tpc_finish(trans)

            for tmp_blob in tmp_blobs_to_rm:
                logger.debug("Removing temporary blob file %s", tmp_blob)
                try:
                    os.unlink(tmp_blob)
                except OSError:
                    pass
            del tmp_blobs_to_rm[:]

            if txnum % self.log_count == 0 and time.time() > log_at:
                now = time.time()
                log_at = now + self.log_interval

                pct_complete = '%1.2f%%' % (txnum * 100.0 / num_txns)
                elapsed = now - begin_time
                if elapsed:
                    rate = total_size / 1e6 / elapsed
                else:
                    rate = 0.0
                rate_str = '%1.3f' % rate

                logger.info(
                    "Copied tid %d,%5d records | %6s MB/s (%6d/%6d,%7s)",
                    bytes8_to_int64(trans.tid), num_txn_records, rate_str,
                    txnum, num_txns, pct_complete)

        elapsed = time.time() - begin_time
        logger.info(
            "All %d transactions copied successfully in %4.1f minutes.",
            txnum, elapsed / 60.0)



##############################################################################
#
# Copyright (c) 2009,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

import os
import time

import zc.lockfile

def lock_blob(path, retries=6000):
    lockfilename = os.path.join(os.path.dirname(path), '.lock')
    n = 0
    while 1:
        try:
            return zc.lockfile.LockFile(lockfilename)
        except zc.lockfile.LockError:
            n += 1
            if n > retries:
                raise
            time.sleep(0.01)



# -*- coding: utf-8 -*-
"""
SQL data types.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function



class Type(object):
    """
    A database type.
    """

    def to_sql_datatype(self, datatype_map):
        return datatype_map[type(self)]

class Unknown(Type):
    "Unspecified."

class Integer64(Type):
    """
    A 64-bit integer.
    """

class UnsignedInteger64(Integer64):
    """
    A 64-bit unsigned integer.
    """

class OID(UnsignedInteger64):
    """
    Type of an OID.
    """

class TID(UnsignedInteger64):
    """
    Type of a TID.
    """

class BinaryString(Type):
    """
    Arbitrary sized binary string.
    """

class State(Type):
    """
    Used for storing object state.
    """

class Boolean(Type):
    """
    A two-value column.
    """

class Char(Type):
    def __init__(self, char_count):
        self.char_count = char_count

    def to_sql_datatype(self, datatype_map):
        return 'CHAR(%s)' % (self.char_count,)



# -*- coding: utf-8 -*-
"""
The ``INSERT`` statement.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import implementer

from .query import Query
from ._util import copy
from .query import ColumnList
from .ast import resolved_against

from .interfaces import ITypedParams
from .interfaces import IOrderedBindParam
from .query import WhereMixin
from .expressions import AssignmentExpression
from .expressions import EmptyExpression
from .expressions import Expression

class _ValuesPlaceholderList(ColumnList):
    pass

@implementer(ITypedParams)
class Insert(Query):

    column_list = None
    select = None
    epilogue = ''
#    values = None

    def __init__(self, table, *columns):
        super(Insert, self).__init__()
        self.table = table

        if columns:
            self.column_list = ColumnList(resolved_against(columns, table))

    def from_select(self, names, select):
        i = copy(self)
        i.column_list = ColumnList(resolved_against(names, select))
        i.select = select
        return i

    def _visit_command(self, compiler):
        compiler.emit_keyword_insert_into()

    def _visit_select(self, compiler):
        compiler.visit(self.select)

    def __compile_visit__(self, compiler):
        self._visit_command(compiler)
        compiler.visit(self.table)
        if self.column_list:
            compiler.visit_grouped(self.column_list)
        if self.select:
            self._visit_select(compiler)
        else:
            compiler.emit_keyword('VALUES')
            if self.column_list:
                values = _ValuesPlaceholderList([self.orderedbindparam()
                                                 for _ in self.column_list])
                compiler.visit_grouped(values)
            else:
                compiler.visit_no_values()
        compiler.emit(self.epilogue)

    def datatypes_for_parameters(self):
        dialect = self.dialect
        if self.column_list and not self.select:
            # If we're sending in a list of values, those have to
            # exactly match the columns, so we can easily get a list
            # of datatypes.
            column_list = self.column_list
            return dialect.datatypes_for_columns(column_list)

        if self.select and self.select.column_list.has_bind_param():
            targets = self.column_list
            sources = self.select.column_list
            # TODO: This doesn't support bind params anywhere except the
            # select list!
            # TODO: This doesn't support named bind params.
            columns_with_params = [
                target
                for target, source in zip(targets, sources)
                if IOrderedBindParam.providedBy(source)
            ]
            return dialect.datatypes_for_columns(columns_with_params)


class _ExcludedColumn(Expression):

    def __init__(self, name):
        self.name = name

    def __compile_visit__(self, compiler): # pragma: no cover
        raise AssertionError("Should only be used in upsert")

    def __compile_visit_for_upsert__(self, compiler):
        compiler.visit_upsert_excluded_column(self)

class Upsert(Insert):
    """
    Perform an insert-or-update operation.

    All supported databases have some version of this,
    but not all of them have the same expressive power. This
    interface is limited to the lowest common denominator.

    You must call ``on_conflict()`` and ``do_update()``.
    The on_conflict parameter must be a single column of the
    primary key, though this isn't verified. ``do_update``
    should be called to specify a subset of the columns in the
    insert list to be updated; they will have the same values as the insert
    list. Note that some databases may update all columns.
    """

    conflict_column = None
    update_columns = None

    def on_conflict(self, exp):
        i = copy(self)
        i.conflict_column = exp
        return i

    def do_update(self, *columns):
        update_columns = ColumnList(resolved_against(columns, self.table))
        assert update_columns.is_subset(self.column_list)
        i = copy(self)
        i.update_columns = update_columns
        return i

    @property
    def update_clause(self):
        return Update(EmptyExpression(), [
            AssignmentExpression(col, _ExcludedColumn(col.name))
            for col in self.update_columns # pylint:disable=not-an-iterable
        ])

    def _visit_command(self, compiler):
        compiler.emit_keyword_upsert()

    def _visit_select(self, compiler):
        super(Upsert, self)._visit_select(compiler)
        compiler.visit_upsert_after_select(self.select)

    def __compile_visit__(self, compiler):
        assert self.conflict_column, "Didn't call on_conflict"
        assert self.update_columns, "Didn't call do_update"
        with compiler.visiting_upsert(self) as upsert_compiler:
            upsert_compiler.visit_upsert(self)

    def __compile_visit_for_upsert__(self, compiler):
        super(Upsert, self).__compile_visit__(compiler)
        compiler.visit_upsert_conflict_column(self.conflict_column)
        compiler.visit_upsert_conflict_update(self.update_clause)

class Insertable(object):

    def insert(self, *columns):
        return Insert(self, *columns)


class Delete(Query, WhereMixin):

    _limit = None

    def __init__(self, table):
        super(Delete, self).__init__()
        self.table = table
        self._where = None

    def limit(self, literal):
        s = copy(self)
        s._limit = literal
        return s

    def __compile_visit__(self, compiler):
        compiler.emit_keyword('DELETE FROM')
        compiler.visit(self.table)
        if self._where:
            compiler.visit(self._where)
        compiler.visit_limit(self._limit)

class Truncate(Query):

    def __init__(self, table):
        super(Truncate, self).__init__()
        self.table = table

    def __compile_visit__(self, compiler):
        compiler.emit_keyword_truncate_table()
        compiler.visit(self.table)


class Deletable(object):
    def delete(self):
        return Delete(self)

    def truncate(self):
        return Truncate(self)

class Updatable(object):
    c = None

    def update(self, **kwargs):
        """
        Update the table. The kwargs must name columns that are members of this table.
        """
        col_expressions = []
        for k, v in kwargs.items():
            col_expressions.append(AssignmentExpression(
                getattr(self.c, k),
                v
            ))

        return Update(self, col_expressions)

class Update(Query, WhereMixin):

    def __init__(self, table, col_expressions):
        self.table = table
        self.col_expressions = col_expressions

    def __compile_visit__(self, compiler):
        compiler.emit_keyword('UPDATE')
        compiler.visit(self.table)
        compiler.emit_keyword('SET')
        compiler.visit_csv(self.col_expressions)
        compiler.visit(self._where)


class Upsertable(object):

    def upsert(self, *columns):
        return Upsert(self, *columns)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..txncontrol import GenericTransactionControl


class Sqlite3TransactionControl(GenericTransactionControl):

    def commit_phase2(self, store_connection, txn, load_connection):
        # When committing, terminate the load connection's transaction now.
        # This allows any actions taken on commit, such as SQLite's auto-checkpoint,
        # to see a state where this reader is not holding open old MVCC resources.
        # See https://github.com/zodb/relstorage/issues/401
        load_connection.rollback_quietly()
        GenericTransactionControl.commit_phase2(self, store_connection, txn, load_connection)



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Wrappers for IStateCache.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope import interface

from relstorage.cache.interfaces import IStateCache

@interface.implementer(IStateCache)
class MultiStateCache(object):
    """
    A proxy that encapsulates the handling of a local and a global
    cache.

    This lets us write loop-free code in all cases in StorageCache,
    but still use both local and global caches.

    For the case that there's just a local cache, which is common and
    semi-recommended, we can just use that object directly.
    """

    __slots__ = ('l', 'g')

    def __init__(self, lcache, gcache):
        self.l = lcache
        self.g = gcache

    def close(self):
        if self.l is not None:
            self.l.close()
            self.g.close()
            self.l = None
            self.g = None

    def release(self):
        if self.l is not None:
            self.l.release()
            self.g.release()
            self.l = None
            self.g = None

    def new_instance(self):
        return type(self)(self.l.new_instance(), self.g.new_instance())

    def flush_all(self):
        self.l.flush_all()
        self.g.flush_all()

    def __contains__(self, key):
        # This has race conditions, but that's ok.
        return key in self.l or key in self.g

    def __getitem__(self, key):
        result = self.l[key]
        if not result:
            result = self.g[key]
            if result:
                self.l[key] = result
        return result

    def get(self, key, peek=False):
        result = self.l.get(key, peek)
        if not result:
            result = self.g.get(key, peek)
            if result:
                self.l[key] = result
        return result

    def __setitem__(self, key, value):
        self.l[key] = value
        self.g[key] = value

    def __delitem__(self, key):
        del self.l[key]
        del self.g[key]

    def set_all_for_tid(self, tid_int, state_oid_iter):
        # If the transaction was very large, it may have spilled
        # over to disk, in which case we will need to read it
        # again (but the second time it's probably at least in the OS's
        # cache). Unless it was very, very large, it probably all winds
        # up in our local cache; we could avoid the second disk read by
        # making a note of (oid, tid) pairs and then pulling them back out
        # of the local cache. (Indeed, simply calling list() to materialize
        # the iterator is probably usually sufficient, except for those very, very
        # large cases.)
        self.l.set_all_for_tid(tid_int, state_oid_iter)
        self.g.set_all_for_tid(tid_int, state_oid_iter)

    def invalidate_all(self, oids):
        self.l.invalidate_all(oids)
        self.g.invalidate_all(oids)


@interface.implementer(IStateCache)
class TracingStateCache(object):
    """
    An ``IStateCache`` implementation that handles tracing.`
    """
    __slots__ = ('cache', 'tracer', '_trace', '_trace_store_current')

    def __init__(self, cache, tracer):
        self.cache = cache
        self.tracer = tracer
        self._trace = tracer.trace
        self._trace_store_current = tracer.trace_store_current

    def new_instance(self):
        return type(self)(self.cache.new_instance(), self.tracer)

    def release(self):
        if self.cache is not None:
            self.cache.release()
            self.cache = None
            # But keep self.tracer because it has to be closed.
            self._trace = None
            self._trace_store_current = None

    def close(self):
        tracer = self.tracer
        self.release()
        if tracer is not None:
            tracer.close()
        self.tracer = None

    def __getattr__(self, name):
        return getattr(self.cache, name)

    def __getitem__(self, key, peek=False):
        oid_int, tid_int = key
        cache_data = self.cache.get(key, peek)
        if cache_data:
            # Note that we trace all cache hits, not just the local cache hit.
            # This makes the simulation less useful, but the stats might still have
            # value to people trying different tuning options manually.
            self._trace(0x22, oid_int, tid_int, dlen=len(cache_data[0]))
        else:
            self._trace(0x20, oid_int)
        return cache_data

    get = __getitem__

    def __setitem__(self, key, value):
        oid_int, _ = key
        state, tid_int = value
        # Record this as a store into the cache, ZEO does.
        self._trace(0x52, oid_int, tid_int, dlen=len(state) if state else 0)
        self.cache[key] = value

    def __delitem__(self, key):
        # TODO: Figure out an event for an explicit invalidation
        del self.cache[key]

    def set_all_for_tid(self, tid_int, state_oid_iter):
        self._trace_store_current(tid_int, state_oid_iter)
        self.cache.set_all_for_tid(tid_int, state_oid_iter)



class _MapWrapper(object): # pragma: no cover
    # This was used for updating the old "checkpoint" style maps
    # but we don't use those anymore. Naively dropping this into
    # mvcc._vacuum() around obsolete_bucket produced unreadable
    # trace files --- but that's probably an artifact of our
    # artifical test, or the artificial way we implemented
    # __delitem__

    # Every setitem generates 0x1c invalidate (hit, saving non-current)
    __slots__ = ('data', 'get', 'trace')
    def __init__(self, data, trace):
        self.get = data.get
        self.trace = trace
        self.data = data

    def __setitem__(self, oid_int, tid_int):
        self.trace(0x1C, oid_int, tid_int)
        self.data[oid_int] = tid_int



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""The core of RelStorage, a ZODB storage for relational databases.

Stores pickles in the database.
"""
from __future__ import absolute_import
from __future__ import print_function

from ZODB.POSException import ReadConflictError
from ZODB.POSException import StorageTransactionError
from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64
from ZODB.utils import z64 as NO_PREV_TID


from relstorage._compat import base64_decodebytes
from relstorage._compat import dumps
from relstorage._compat import metricmethod_sampled
from relstorage._compat import OID_TID_MAP_TYPE
from relstorage._util import to_utf8

from . import AbstractTPCStateDatabaseAvailable
from .vote import DatabaseLockedForTid
from .vote import HistoryFree as HFVoteFactory
from .vote import HistoryPreserving as HPVoteFactory
from .vote import HistoryPreservingDeleteOnly as HPDeleteOnlyVoteFactory


class TransactionConflictsWithItselfError(ReadConflictError):
    """
    The transaction asked for two different OIDs for the same object.
    """

class _BadFactory(object):

    def enter(self, storage):
        raise NotImplementedError

class AbstractBegin(AbstractTPCStateDatabaseAvailable):
    """
    The phase we enter after ``tpc_begin`` has been called.
    """
    __slots__ = (
        # (user, description, extension) from the transaction.
        # byte objects.
        'ude',
        # The location of the temporary data.
        'temp_storage',

        # required_tids: {oid_int: tid_int}; confirms that certain objects
        # have not changed at commit. May be a BTree
        'required_tids',

        # The factory we call to produce a voting state. Must return
        # an object with an enter() method.
        'tpc_vote_factory',

        # OIDs of things we have deleted or undone.
        # Stored in their 8 byte form
        'invalidated_oids',

    )

    _DEFAULT_TPC_VOTE_FACTORY = _BadFactory # type: Callable[..., AbstractTPCState]

    def __init__(self, shared_state):
        super(AbstractBegin, self).__init__(shared_state)
        self.invalidated_oids = ()
        # We'll replace this later with the right type when it's needed.
        self.required_tids = {} # type: Dict[int, int]
        self.tpc_vote_factory = self._DEFAULT_TPC_VOTE_FACTORY # type: ignore

        user = to_utf8(self.transaction.user)
        desc = to_utf8(self.transaction.description)
        ext = self.transaction.extension

        if ext:
            ext = dumps(ext, 1)
        else:
            ext = b""
        self.ude = user, desc, ext

    def tpc_vote(self, storage, transaction):
        if transaction is not self.transaction:
            raise StorageTransactionError(
                "tpc_vote called with wrong transaction")

        next_phase = self.tpc_vote_factory(self)
        next_phase.enter(storage)
        return next_phase

    def store(self, oid, previous_tid, data, transaction):
        """
        This method should take no globally visible commit locks.
        """
        # Called by Connection.commit(), after tpc_begin has been called.
        if transaction is not self.transaction:
            raise StorageTransactionError(self, transaction)

        oid_int = bytes8_to_int64(oid)
        if previous_tid and previous_tid != NO_PREV_TID:
            # previous_tid is the tid of the state that the
            # object was loaded from. cPersistent objects return a brand
            # new bytes object each time even if it's all zeros; Python implementation
            # returns a private constant. It would be nice if they all returned a public
            # interned constant so we could compare with `is`.
            prev_tid_int = bytes8_to_int64(previous_tid)
        else:
            prev_tid_int = 0

        # Save the data locally in a temporary place. Later, closer to commit time,
        # we'll send it all over at once. This lets us do things like use
        # COPY in postgres.
        self.shared_state.temp_storage.store_temp(oid_int, data, prev_tid_int)

    @metricmethod_sampled
    def checkCurrentSerialInTransaction(self, oid, required_tid, transaction):
        if transaction is not self.transaction:
            raise StorageTransactionError(self, transaction)

        required_tid_int = bytes8_to_int64(required_tid)
        oid_int = bytes8_to_int64(oid)

        # If this transaction already specified a different serial for
        # this oid, the transaction conflicts with itself.
        required_tids = self.required_tids
        if not required_tids:
            required_tids = self.required_tids = OID_TID_MAP_TYPE()

        previous_serial_int = required_tids.get(oid_int, required_tid_int)
        if previous_serial_int != required_tid_int:
            raise TransactionConflictsWithItselfError(
                oid=oid,
                serials=(int64_to_8bytes(previous_serial_int),
                         required_tid))
        required_tids[oid_int] = required_tid_int

        # Previously, we used Loader.getTid() (a wrapper around
        # Loader.load()) to check that what's visible in the database
        # load connection is what they asked for.
        #
        # But this shouldn't actually matter: This object came from
        # the load connection in the first place. This is probably a
        # cache hit. If it doesn't match, then we failed to notice a
        # change and invalidate the object. This is called from
        # Connection.commit(), which happens after tpc_begin() and
        # just before tpc_vote(), so there's no real reason not to
        # just defer it a bit until we do it in bulk for vote.

    def _invalidated_oids(self, *oid_bytes):
        if not self.invalidated_oids:
            self.invalidated_oids = set()
        self.invalidated_oids.update(oid_bytes)


    def deleteObject(self, oid, oldserial, transaction):
        """
        This method operates directly against the ``object_state`` table;
        as such, it immediately takes out locks on that table.

        This method is only expected to be called when performing
        ``IExternalGC`` operations (e.g., from zc.zodbdgc
        or from ZODB/tests/IExternalGC.test).

        In history-free mode, deleting objects does not allocate
        a new tid (well, it allocates it, but there's no place to store
        it). In history preserving mode, it will wind up allocating a tid
        to store the empty transaction (only previous states were undone)

        TODO: This needs a better, staged implementation. I think it is
           highly likely to deadlock now if anything happened to be reading
           those rows.
        XXX: If we have blobs in a non-shared disk location, this does not
           remove them.
        """
        if transaction is not self.transaction: # pragma: no cover
            raise StorageTransactionError(self, transaction)

        # We shouldn't have to worry about anything in self._cache
        # because by definition we are deleting objects that were not
        # reachable and so shouldn't be in the cache (or if they were,
        # we'll never ask for them anyway). Most likely, this is running
        # in a separate process anyway, not used for regular storage (
        # an instance of multi-zodb-gc). However, in case it is in a regular
        # process, and in case we do have other transactions that could theoretically
        # see this state, and to relieve memory pressure on local/global caches,
        # we do go ahead and invalidate a cached entry.
        # TODO: We need a distinct name for invalidate, so we can differentiate
        # between why we're doing it. Did we write a newer version? Did we
        # delete a specific verison? Etc.
        oid_int = bytes8_to_int64(oid)
        tid_int = bytes8_to_int64(oldserial)
        self.shared_state.cache.remove_cached_data(oid_int, tid_int)

        # We delegate the actual operation to the adapter's packundo,
        # just like native pack
        cursor = self.shared_state.store_connection.cursor
        # When this is done, we get a tpc_vote,
        # and a tpc_finish.
        # The interface doesn't specify a return value, so for testing
        # we return the count of rows deleted (should be 1 if successful)
        deleted = self.shared_state.adapter.packundo.deleteObject(cursor, oid, oldserial)
        self._invalidated_oids(oid)
        return deleted


class HistoryFree(AbstractBegin):

    __slots__ = ()

    _DEFAULT_TPC_VOTE_FACTORY = HFVoteFactory


class HistoryPreserving(AbstractBegin):

    __slots__ = (
        # If we use undo() or deleteObject(), we have to allocate a TID, which means
        # we have to lock the database. Not cool.
        'committing_tid_lock',
    )

    _DEFAULT_TPC_VOTE_FACTORY = HPVoteFactory

    def __init__(self, *args):
        AbstractBegin.__init__(self, *args)
        self.committing_tid_lock = None

    def _obtain_commit_lock(self, cursor):
        if self.committing_tid_lock is None:
            # Note that _prepare_tid acquires the commit lock.
            # The commit lock must be acquired after the pack lock
            # because the database adapters also acquire in that
            # order during packing.
            tid_lock = DatabaseLockedForTid.lock_database_for_next_tid(
                cursor, self.shared_state.adapter, self.ude)
            self.committing_tid_lock = tid_lock

    def deleteObject(self, oid, oldserial, transaction):
        # XXX: Maybe we don't need the commit lock at all? Since
        # theoretically these are unreachable? Our custom
        # vote stage just removes this transaction anyway; maybe it
        # can skip the committing.
        self._obtain_commit_lock(self.shared_state.store_connection.cursor)
        # A transaction that deletes objects can *only* delete objects.
        # That way we don't need to store an entry in the transaction table
        # (and add extra bloat to the DB; that kind of defeats the point of
        # deleting objects).
        self.tpc_vote_factory = HPDeleteOnlyVoteFactory
        return super(HistoryPreserving, self).deleteObject(oid, oldserial, transaction)

    def undo(self, transaction_id, transaction):
        """
        This method temporarily holds the pack lock, releasing it when
        done, and it also holds the commit lock, keeping it held for
        the next phase.

        Returns an iterable of ``(oid_int, tid_int)`` pairs giving the
        items that were restored and are now current. All of those oids that
        had any data stored for ``transaction_id`` are now invalid.
        """
        # Typically if this is called, the store/restore methods will *not* be
        # called, but there's not a strict guarantee about that.
        if transaction is not self.transaction:
            raise StorageTransactionError(self, transaction)

        # Unlike most places, transaction_id is the base 64 encoding
        # of an 8 byte tid

        undo_tid = base64_decodebytes(transaction_id + b'\n') # pylint:disable=deprecated-method
        assert len(undo_tid) == 8
        undo_tid_int = bytes8_to_int64(undo_tid)

        adapter = self.shared_state.adapter
        cursor = self.shared_state.store_connection.cursor
        assert cursor is not None

        adapter.locker.hold_pack_lock(cursor)
        try:
            adapter.packundo.verify_undoable(cursor, undo_tid_int)
            self._obtain_commit_lock(cursor)

            self_tid_int = self.committing_tid_lock.tid_int
            copied = adapter.packundo.undo(
                cursor, undo_tid_int, self_tid_int)

            # Invalidate all cached data for these oids. We have a
            # brand new transaction ID that's greater than any they
            # had before. In history-preserving mode, there could
            # still be other valid versions. See notes in packundo:
            # In theory we could be undoing a transaction several generations in the
            # past where the object had multiple intermediate states, but in practice
            # we're probably just undoing the latest state. Still, play it
            # a bit safer.
            oid_ints = [oid_int for oid_int, _ in copied]
            self.shared_state.cache.remove_all_cached_data_for_oids(oid_ints)

            # Update the current object pointers immediately, so that
            # subsequent undo operations within this transaction will see
            # the new current objects.
            adapter.mover.update_current(cursor, self_tid_int)

            self.shared_state.blobhelper.copy_undone(copied,
                                                     self.committing_tid_lock.tid)

            oids = [int64_to_8bytes(oid_int) for oid_int in oid_ints]
            self._invalidated_oids(*oids)

            return copied
        finally:
            adapter.locker.release_pack_lock(cursor)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
IOIDAllocator implementations.
"""

from __future__ import absolute_import

import abc
from contextlib import contextmanager

from relstorage._compat import metricmethod
from relstorage._compat import ABC

from .schema import Schema

logger = __import__('logging').getLogger(__name__)

class AbstractOIDAllocator(ABC):
    """
    These objects are intended to be stateless.

    They consult an underlying database sequence than increases, and
    then derive additional OIDs from that sequence. ``new_oids``
    returns a list of those OIDs, with the largest such OID in
    position 0.
    """
    __slots__ = ()

    def new_instance(self):
        return self

    close = release = lambda self: None

    def reset_oid(self, cursor):
        raise NotImplementedError

    @abc.abstractmethod
    def set_min_oid(self, cursor, oid_int):
        raise NotImplementedError

    @abc.abstractmethod
    def new_oids(self, cursor):
        raise NotImplementedError

# All of these allocators allocate 16 OIDs at a time. In the sequence
# or table, value (n) represents (n * 16 - 15) through (n * 16). So,
# value 1 represents OID block 1-16, 2 represents OID block 17-32, and
# so on. The _oid_range_around helper method returns a list around
# this number sorted in the proper way (largest to smallest, so that
# list.pop() returns the smallest remaining number, and so on:
# [16, 15, ..., 1] when given 1)
#
# Given:
#    num_oids = 16
#    highest_inclusive = n * num_oids2
#    highest_exclusive = highest_inclusive - 1
#    lowest_inclusive = highest_inclusive - (num_oids - 1)
#    lowest_exclusive = lowest_inclusive - 1

# Note that
#    range(lowest_inclusive, highest_inclusive + 1).sort(reverse=True)
#
# is the same as
#     range(highest_inclusive, lowest_exclusive, -1)

_OID_RANGE_SIZE = 16
def _oid_range_around_assume_list(n, _s=_OID_RANGE_SIZE):
    return range(n * _s, n * _s - _s, -1)

def _oid_range_around_iterable(n, _s=_OID_RANGE_SIZE, _range=_oid_range_around_assume_list):
    return list(_range(n))


class AbstractRangedOIDAllocator(AbstractOIDAllocator):
    __slots__ = ()

    def set_min_oid(self, cursor, oid_int):
        # This takes a user-space OID and turns it into the internal
        # range number.
        n = (oid_int + 15) // 16
        self._set_min_oid_from_range(cursor, n)

    @abc.abstractmethod
    def _set_min_oid_from_range(self, cursor, n):
        raise NotImplementedError

    if isinstance(range(1), list):
        # Py2
        _oid_range_around = staticmethod(_oid_range_around_assume_list)
    else:
        _oid_range_around = staticmethod(_oid_range_around_iterable)


class AbstractTableOIDAllocator(AbstractRangedOIDAllocator):
    """
    Implements OID allocation using a table, new_oid, with an incrementing
    column.
    """
    # pylint:disable=abstract-method

    #: To use bound queries, we need to know whether we're keeping history.
    #: It doesn't change anything, though.
    keep_history = None

    #: After this many allocated OIDs should the (unlucky) thread that
    #: allocated the one evenly divisible by this number attempt to remove
    #: old OIDs. Set to None to disable periodic garbage collection.
    garbage_collect_interval = 100001

    # How many OIDs to attempt to delete at any one request. Keeping
    # this on the small side relative to the interval limits the
    # chances of deadlocks (by observation).
    garbage_collect_interactive_size = 1000

    # How many OIDs to attempt to delete if we're not allocating, only
    # garbage collecting.
    garbage_collect_batch_size = 3000

    # How long to wait in seconds before timing out when batch deleting locks.
    # Note that we don't do this interactively because our connection
    # already has a custom timeout.
    garbage_collect_batch_timeout = 5

    def __init__(self, driver):
        # Take the driver so we can bind to our queries
        self.driver = driver

    _insert_stmt = Schema.new_oid.insert()

    @metricmethod
    def new_oids(self, cursor):
        """Return a sequence of new, unused OIDs."""
        self._insert_stmt.execute(cursor)

        # This is a DB-API extension. Fortunately, all
        # supported drivers implement it.
        n = cursor.lastrowid

        if self.garbage_collect_interval and n % self.garbage_collect_interval == 0:
            self.garbage_collect_oids(cursor, n)
        return self._oid_range_around(n)

    @contextmanager
    def _timeout(self, cursor, batch_timeout):
        yield (cursor, batch_timeout)

    def garbage_collect_oids(self, cursor, max_value=None):
        # Clean out previously generated OIDs.
        batch_size = self.garbage_collect_interactive_size
        batch_timeout = None
        if not max_value:
            batch_size = self.garbage_collect_batch_size
            batch_timeout = self.garbage_collect_batch_timeout
            cursor.execute('SELECT MAX(zoid) FROM new_oid')
            row = cursor.fetchone()
            if row:
                max_value = row[0]
        if not max_value:
            return
        # Delete old OIDs, starting with the oldest.
        stmt = """
        DELETE IGNORE
        FROM new_oid
        WHERE zoid < %s
        ORDER BY zoid ASC
        LIMIT %s
        """
        params = (max_value, batch_size)
        __traceback_info__ = stmt, params
        rowcount = True
        with self._timeout(cursor, batch_timeout):
            while rowcount:
                try:
                    cursor.execute(stmt, params)
                except Exception: # pylint:disable=broad-except
                    # Luckily, a deadlock only rolls back the previous
                    # statement, not the whole transaction.
                    #
                    # XXX: No, that's not true. A general error, like
                    # a lock timeout, will roll back the previous
                    # statement. A deadlock rolls back the whole
                    # transaction. We're lucky the difference here
                    # doesn't make any difference: we don't actually
                    # write anything to the database temp tables, etc,
                    # until the storage enters commit(), at which
                    # point we shouldn't need to allocate any more new
                    # oids.
                    #
                    # TODO: We'd prefer to only do this for errcode 1213: Deadlock.
                    # MySQLdb raises this as an OperationalError; what do all the other
                    # drivers do?
                    logger.debug("Failed to garbage collect allocated OIDs", exc_info=True)
                    break
                else:
                    rowcount = cursor.rowcount
                    logger.debug("Garbage collected %s old OIDs less than %s", rowcount, max_value)

    _reset_oid_query = Schema.new_oid.truncate()

    def reset_oid(self, cursor):
        self._reset_oid_query.execute(cursor)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for the SQL abstraction layer.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase

from .. import Table
from .. import TemporaryTable
from .. import Char
from .. import HistoryVariantTable
from .. import Column
from .. import it
from .. import DefaultDialect
from .. import OID
from .. import TID
from .. import State
from .. import Boolean
from .. import BinaryString
from .. import func

from ..expressions import bindparam

current_object = Table(
    'current_object',
    Column('zoid', OID),
    Column('tid', TID)
)

object_state = Table(
    'object_state',
    Column('zoid', OID),
    Column('tid', TID),
    Column('state', State),
    Column('state_size'),
)

hp_object_and_state = current_object.natural_join(object_state)

objects = HistoryVariantTable(
    current_object,
    object_state,
)

object_and_state = HistoryVariantTable(
    hp_object_and_state,
    object_state
)

transaction = Table(
    'transaction',
    Column('tid', TID),
    Column('packed', Boolean),
    Column('username', BinaryString),
    Column('description', BinaryString),
    Column('extension', BinaryString),
)

temp_store = TemporaryTable(
    'temp_store',
    Column('zoid', OID, primary_key=True),
    Column('prev_tid', TID, nullable=False),
    Column('md5', Char(32)), # Only used when keep_history=True
    Column('state', State)
)


class TestTableSelect(TestCase):

    def test_simple_eq_select(self):
        table = object_state

        stmt = table.select().where(table.c.zoid == table.c.tid)

        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size FROM object_state WHERE (zoid = tid)'
        )

    def test_simple_eq_limit(self):
        table = object_state

        stmt = table.select().where(table.c.zoid == table.c.tid).limit(1)

        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size FROM object_state WHERE (zoid = tid) LIMIT 1'
        )

    def test_simple_eq_for_update(self):
        table = object_state

        stmt = table.select().where(table.c.zoid == table.c.tid).for_update()

        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size FROM object_state WHERE (zoid = tid) FOR UPDATE'
        )

        stmt = stmt.nowait()
        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size FROM object_state WHERE (zoid = tid) FOR UPDATE '
            'NOWAIT'
        )

    def test_max_select(self):
        table = object_state

        stmt = table.select(func.max(it.c.tid))

        self.assertEqual(
            str(stmt),
            'SELECT max(tid) FROM object_state'
        )

    def test_distinct(self):
        table = object_state
        stmt = table.select(table.c.zoid).where(table.c.tid == table.bindparam('tid')).distinct()
        self.assertEqual(
            str(stmt),
            'SELECT DISTINCT zoid FROM object_state WHERE (tid = %(tid)s)'
        )

    def test_simple_eq_select_and(self):

        table = object_state

        stmt = table.select().where(table.c.zoid == table.c.tid)

        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size FROM object_state WHERE (zoid = tid)'
        )

        stmt = stmt.and_(table.c.zoid > 5)
        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size '
            'FROM object_state WHERE ((zoid = tid AND zoid > 5))'
        )

    def test_simple_eq_select_literal(self):
        table = object_state

        # This is a useless query
        stmt = table.select().where(table.c.zoid == 7)

        self.assertEqual(
            str(stmt),
            'SELECT zoid, tid, state, state_size FROM object_state WHERE (zoid = 7)'
        )

        self.assertEqual(
            stmt.compiled().params,
            {})

    def test_column_query_variant_table(self):
        stmt = objects.select(objects.c.tid, objects.c.zoid).where(
            objects.c.tid > bindparam('tid')
        )

        self.assertEqual(
            str(stmt),
            'SELECT tid, zoid FROM current_object WHERE (tid > %(tid)s)'
        )

    def test_natural_join(self):
        stmt = object_and_state.select(
            object_and_state.c.zoid, object_and_state.c.state
        ).where(
            object_and_state.c.zoid == object_and_state.bindparam('oid')
        )

        self.assertEqual(
            str(stmt),
            'SELECT zoid, state '
            'FROM current_object '
            'JOIN object_state '
            'USING (tid, zoid) WHERE (zoid = %(oid)s)'
        )

        class H(object):
            keep_history = False
            dialect = DefaultDialect()

        stmt = stmt.bind(H())

        self.assertEqual(
            str(stmt),
            'SELECT zoid, state '
            'FROM object_state '
            'WHERE (zoid = %(oid)s)'
        )

    def test_bind(self):
        from operator import attrgetter
        # pylint:disable=no-member
        select = objects.select(objects.c.tid, objects.c.zoid).where(
            objects.c.tid > bindparam('tid')
        )
        # Unbound we assume history
        self.assertEqual(
            str(select),
            'SELECT tid, zoid FROM current_object WHERE (tid > %(tid)s)'
        )

        class Context(object):
            dialect = DefaultDialect()
            keep_history = True

        context = Context()
        dialect = context.dialect
        query = select.bind(context)

        class Root(object):
            select = query

        # The table replaced itself with the correct variant.
        self.assertIsNot(query.table, objects)
        self.assertIs(query.table, objects.history_preserving)

        for item_name in (
                'select',
                'select._where',
                'select._where.expression',
        ):
            __traceback_info__ = item_name
            item = attrgetter(item_name)(Root)
            # The exact context is passed down the tree.
            self.assertIs(item.context, context)
            # The dialect is first bound, so it's *not* the same
            # as the one we can reference (though it is equal)...
            self.assertEqual(item.dialect, dialect)
            # ...but it *is* the same throughout the tree
            self.assertIs(query.dialect, item.dialect)

        # We take up its history setting
        self.assertEqual(
            str(query),
            'SELECT tid, zoid FROM current_object WHERE (tid > %(tid)s)'
        )

        # Bound to history-free we use history free
        context.keep_history = False
        query = select.bind(context)

        self.assertEqual(
            str(query),
            'SELECT tid, zoid FROM object_state WHERE (tid > %(tid)s)'
        )

    def test_bind_descriptor(self):
        class Context(object):
            keep_history = True
            dialect = DefaultDialect()
            # pylint:disable=no-member
            select = objects.select(objects.c.tid, objects.c.zoid).where(
                objects.c.tid > bindparam('tid')
            )

        # Unbound we assume history
        self.assertEqual(
            str(Context.select),
            'SELECT tid, zoid FROM current_object WHERE (tid > %(tid)s)'
        )

        context = Context()
        context.keep_history = False
        self.assertEqual(
            str(context.select),
            'SELECT tid, zoid FROM object_state WHERE (tid > %(tid)s)'
        )

    def test_prepared_insert_values(self):
        stmt = current_object.insert(
            current_object.c.zoid
        )
        stmt.__name__ = 'ins'

        self.assertEqual(
            str(stmt),
            'INSERT INTO current_object(zoid) VALUES (%s)'
        )

        stmt = stmt.prepared()
        self.assertTrue(
            str(stmt).startswith('EXECUTE rs_prep_stmt')
        )

        stmt = stmt.compiled()
        self.assertRegex(
            stmt._prepare_stmt,
            r"PREPARE rs_prep_stmt_ins_[0-9]*_[0-9]* \(BIGINT\) AS.*"
        )

    def test_prepared_insert_select_with_param(self):
        stmt = current_object.insert().from_select(
            (current_object.c.zoid,
             current_object.c.tid),
            object_state.select(
                object_state.c.zoid,
                object_state.orderedbindparam()
            )
        )
        stmt.__name__ = 'ins'
        self.assertEqual(
            str(stmt),
            'INSERT INTO current_object(zoid, tid) SELECT zoid, %s FROM object_state'
        )

        stmt = stmt.prepared()
        self.assertTrue(
            str(stmt).startswith('EXECUTE rs_prep_stmt')
        )

        stmt = stmt.compiled()
        self.assertRegex(
            stmt._prepare_stmt,
            r"PREPARE rs_prep_stmt_ins_[0-9]*_[0-9]* \(BIGINT\) AS.*"
        )

    def test_it(self):
        stmt = object_state.select(
            it.c.zoid,
            it.c.state
        ).where(
            it.c.tid == it.bindparam('tid')
        ).order_by(
            it.c.zoid
        )

        self.assertEqual(
            str(stmt),
            'SELECT zoid, state FROM object_state WHERE (tid = %(tid)s) ORDER BY zoid'
        )

        # Now something that won't resolve.
        col_ref = it.c.dne

        # In the column list
        with self.assertRaisesRegex(AttributeError, 'does not include dne'):
            object_state.select(col_ref)

        stmt = object_state.select(it.c.zoid)

        # In the where clause
        with self.assertRaisesRegex(AttributeError, 'does not include dne'):
            stmt.where(col_ref == object_state.c.state)

        # In order by
        with self.assertRaisesRegex(AttributeError, 'does not include dne'):
            stmt.order_by(col_ref == object_state.c.state)

    def test_boolean_literal(self):
        stmt = transaction.select(
            transaction.c.tid
        ).where(
            it.c.packed == False # pylint:disable=singleton-comparison
        ).order_by(
            transaction.c.tid, 'DESC'
        )

        self.assertEqual(
            str(stmt),
            'SELECT tid FROM transaction WHERE (packed = FALSE) ORDER BY tid DESC'
        )

    def test_literal_in_select(self):
        stmt = current_object.select(
            1
        ).where(
            current_object.c.zoid == current_object.bindparam('oid')
        )

        self.assertEqual(
            str(stmt),
            'SELECT 1 FROM current_object WHERE (zoid = %(oid)s)'
        )

    def test_boolean_literal_it_joined_table(self):
        stmt = transaction.natural_join(
            object_state
        ).select(
            it.c.tid, it.c.username, it.c.description, it.c.extension,
            object_state.c.state_size
        ).where(
            it.c.zoid == it.bindparam("oid")
        ).and_(
            it.c.packed == False # pylint:disable=singleton-comparison
        ).order_by(
            it.c.tid, "DESC"
        )

        self.assertEqual(
            str(stmt),
            'SELECT tid, username, description, extension, state_size '
            'FROM transaction '
            'JOIN object_state '
            'USING (tid) '
            'WHERE ((zoid = %(oid)s AND packed = FALSE)) '
            'ORDER BY tid DESC'
        )

class TestUpsert(TestCase):
    # This is here rather than test_dialect.py because it uses the higher level
    # schema objects.
    keep_history = False
    dialect = DefaultDialect()
    maxDiff = None

    insert_or_replace = (
        'INSERT INTO object_state(zoid, state, tid, state_size) '
        'VALUES (%s, %s, %s, %s) '
        'ON CONFLICT (zoid) '
        'DO UPDATE SET state = excluded.state, tid = excluded.tid, '
        'state_size = excluded.state_size'
    )

    def test_insert_or_replace(self):
        stmt = object_state.upsert(
            it.c.zoid,
            it.c.state,
            it.c.tid,
            it.c.state_size
        ).on_conflict(
            it.c.zoid
        ).do_update(
            it.c.state,
            it.c.tid,
            it.c.state_size
        ).bind(self)

        self.assertEqual(
            str(stmt),
            self.insert_or_replace
        )

    insert_or_replace_subquery = (
        'INSERT INTO object_state(zoid, tid, state, state_size) '
        'SELECT zoid, %s, state, COALESCE(LENGTH(state), 0) FROM temp_store '
        'ORDER BY zoid '
        'ON CONFLICT (zoid) '
        'DO UPDATE SET state = excluded.state, tid = excluded.tid, '
        'state_size = excluded.state_size'
    )

    def test_insert_or_replace_subquery(self):
        stmt = object_state.upsert(
            it.c.zoid,
            it.c.tid,
            it.c.state,
            it.c.state_size
        ).from_select(
            (object_state.c.zoid,
             object_state.c.tid,
             object_state.c.state,
             object_state.c.state_size,
            ),
            temp_store.select(
                temp_store.c.zoid,
                temp_store.orderedbindparam(),
                temp_store.c.state,
                'COALESCE(LENGTH(state), 0)',
            ).order_by(
                temp_store.c.zoid
            )
        ).on_conflict(
            it.c.zoid
        ).do_update(
            it.c.state,
            it.c.tid,
            it.c.state_size
        ).bind(self)

        self.maxDiff = None
        self.assertEqual(
            str(stmt),
            self.insert_or_replace_subquery
        )

    upsert_unconstrained_subquery = (
        'INSERT INTO object_state(zoid, tid, state, state_size) '
        'SELECT zoid, %s, state, COALESCE(LENGTH(state), 0) FROM temp_store '
        'ON CONFLICT (zoid) '
        'DO UPDATE SET state = excluded.state, tid = excluded.tid, '
        'state_size = excluded.state_size'
    )

    def test_upsert_unconstrained_subquery(self):
        stmt = object_state.upsert(
            it.c.zoid,
            it.c.tid,
            it.c.state,
            it.c.state_size
        ).from_select(
            (object_state.c.zoid,
             object_state.c.tid,
             object_state.c.state,
             object_state.c.state_size,
            ),
            temp_store.select(
                temp_store.c.zoid,
                temp_store.orderedbindparam(),
                temp_store.c.state,
                'COALESCE(LENGTH(state), 0)',
            )
        ).on_conflict(
            it.c.zoid
        ).do_update(
            it.c.state,
            it.c.tid,
            it.c.state_size
        ).bind(self)

        self.maxDiff = None
        self.assertEqual(
            str(stmt),
            self.upsert_unconstrained_subquery
        )



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Implementation of load methods.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64
from ZODB.utils import maxtid

from relstorage.interfaces import POSKeyError
from relstorage.cache.interfaces import CacheConsistencyError
from .._compat import metricmethod
from .._compat import metricmethod_sampled
from .util import storage_method
from .util import stale_aware


logger = __import__('logging').getLogger(__name__)

def _make_pke_data(cursor, adapter, oid_int, reason):
    """
    Add extra info to the POSKeyError from load.

    KeyErrors in load() are generally not supposed to happen,
    so this is a good place to gather information.
    """
    extra = {'reason': reason}

    if adapter.keep_history:
        tid = adapter.txncontrol.get_tid(cursor)
        extra['current_txn'] = tid

        tids = []
        try:
            history = adapter.dbiter.iter_object_history(cursor, oid_int)
        except KeyError as e:
            # The object has no history, at least from the point of view
            # of the current database load connection.
            tids = str(e)
            del e
        else:
            for entry in history:
                tids.append(entry.tid_int)
                if len(tids) >= 10:
                    break
        extra['recent_tids'] = tids
    return extra

class Loader(object):

    __slots__ = (
        'adapter',
        'load_connection',
        'cache',
        '__dict__',
    )

    def __init__(self, adapter, load_connection, cache):
        self.adapter = adapter
        self.load_connection = load_connection
        self.cache = cache

    def __load_using_method(self, load_cursor, meth, argument):
        try:
            return meth(load_cursor, argument)
        except CacheConsistencyError:
            logger.exception("Cache consistency error; restarting load")
            self.load_connection.drop()
            raise

    def __pke(self, oid_bytes, **extra):
        defs = {
            'cache': self.cache,
            'load': self.load_connection,
            'adapter': self.adapter,
        }
        defs.update(extra)
        return POSKeyError(oid_bytes, defs)

    @stale_aware
    @storage_method
    @metricmethod_sampled
    def load(self, oid, version=''):
        # pylint:disable=unused-argument
        oid_int = bytes8_to_int64(oid)
        # TODO: Here, and in prefetch, should we check bool(load_connection)?
        # If it's not active and had polled, we don't really want to do that, do we?
        load_cursor = self.load_connection.cursor
        state, tid_int = self.__load_using_method(load_cursor, self.cache.load, oid_int)
        if tid_int is None:
            raise self.__pke(oid,
                             **_make_pke_data(load_cursor,
                                              self.adapter,
                                              oid_int,
                                              "no tid found"))

        if not state:
            # This can happen if something attempts to load
            # an object whose creation has been undone or which was deleted
            # by IExternalGC.deleteObject().
            raise self.__pke(oid,
                             **_make_pke_data(load_cursor,
                                              self.adapter,
                                              oid_int,
                                              "creation undone"))
        return state, int64_to_8bytes(tid_int)

    @storage_method
    def getTid(self, oid):
        """
        Return the transaction TID bytes for the OID, as
        seen from the load connection.
        """
        _state, serial = self.load(oid)
        return serial

    @stale_aware
    @storage_method
    def prefetch(self, oids):
        prefetch = self.cache.prefetch
        oid_ints = [bytes8_to_int64(oid) for oid in oids]
        load_cursor = self.load_connection.cursor
        try:
            self.__load_using_method(load_cursor, prefetch, oid_ints)
        except (AttributeError, TypeError, ValueError):
            raise
        except Exception: # pylint:disable=broad-except
            # This could raise self._stale_error, or
            # CacheConsistencyError. Both of those mean that regular loads
            # may fail too, but we don't know what our transaction state is
            # at this time, so we don't want to raise it to the caller.
            logger.exception("Failed to prefetch")

    # This is *NOT* stale aware for some reason (why?)
    @storage_method
    @metricmethod_sampled
    def loadSerial(self, oid, serial):
        """Load a specific revision of an object"""
        oid_int = bytes8_to_int64(oid)
        tid_int = bytes8_to_int64(serial)

        # If we've got this state cached exactly,
        # use it. No need to poll or anything like that first;
        # polling is unlikely to get us the state we want.
        # If the data happens to have been removed from the database,
        # due to a pack, this won't detect it if it was already cached
        # and the pack happened somewhere else. This method is
        # only used for conflict resolution, though, and we
        # shouldn't be able to get to that point if the root revision
        # went missing, right? Packing periodically takes the same locks we
        # want to take for committing.
        state = self.cache.loadSerial(oid_int, tid_int)
        if state:
            return state

        # Allow loading data from later transactions for conflict
        # resolution. There are three states involved in conflict resolution:
        # the original state (our load connection should be able to see that),
        # the committed state (our store connection can see that, and we returned it
        # when we detected the conflict) and the state we're trying to commit
        # (stored in the temporary cache data). So if we get here, mostly we
        # should need to use the load connection.

        state = self.adapter.mover.load_revision(
            self.load_connection.cursor,
            oid_int,
            tid_int)

        if state:
            return state

        # Actually using the store_connection to pull into the future was
        # removed as part of the pooling of store_connection. The above comments
        # indicate that we really shouldn't need to get here, and no tests break
        # with this commented out. What's a legitimate need for pulling into the future?
        # state = self.adapter.mover.load_revision(
        #     self.store_connection.cursor,
        #     oid_int,
        #     tid_int)
        # if state:
        #     return state

        raise self.__pke(oid, tid_int=tid_int, state=state)


    @stale_aware
    @storage_method
    @metricmethod_sampled
    def loadBefore(self, oid, tid):
        """
        Return the most recent revision of oid before tid committed.
        """
        if tid is maxtid or tid == maxtid:
            # This is probably from ZODB.utils.load_current(), which
            # is really trying to just get the current state of the
            # object. This is almost entirely just from test cases; ZODB 5's mvccadapter
            # doesn't even expose it, so ZODB.Connection doesn't use it.
            #
            # Shortcut the logic below by using load() (current),
            # formatted in the way this method returns it:
            #
            #     ``(state, tid # of state, tid_after_state)``
            #
            # where tid_after_state will naturally be None
            return self.load(oid) + (None,)
        oid_int = bytes8_to_int64(oid)

        # TODO: This makes three separate queries, and also bypasses the cache.
        # We should be able to fix at least the multiple queries.

        # In the past, we would use the store connection (only if it was already open)
        # to "allow leading dato from later transactions for conflict resolution".
        # However, this doesn't seem to be used in conflict
        # resolution. ZODB.ConflictResolution.tryToResolveConflict
        # calls loadSerial(); About the only call in ZODB to
        # loadBefore() is from BlobStorage.undo() (which
        # RelStorage does not extend). Mixing and matching calls
        # between connections using different isolation levels
        # isn't great.
        #
        # We had it as a todo for a long time to stop doing that, and
        # pooling store connections was a great time to try it.
        cursor = self.load_connection.cursor
        if not self.adapter.mover.exists(cursor, oid_int):
            raise self.__pke(oid, exists=False)

        state, start_tid = self.adapter.mover.load_before(
            cursor, oid_int, bytes8_to_int64(tid))

        if start_tid is None:
            return None

        if state is None:
            # This can happen if something attempts to load
            # an object whose creation has been undone, see load()
            # This change fixes the test in
            # TransactionalUndoStorage.checkUndoCreationBranch1
            # self._log_keyerror doesn't work here, only in certain states.
            self.__pke(oid, undone=True)

        end_int = self.adapter.mover.get_object_tid_after(
            cursor, oid_int, start_tid)
        if end_int is not None:
            end = int64_to_8bytes(end_int)
        else:
            end = None

        return state, int64_to_8bytes(start_tid), end


class BlobLoader(object):

    __slots__ = (
        'load_connection',
        'blobhelper',
    )

    def __init__(self, load_connection, blobhelper):
        self.load_connection = load_connection
        self.blobhelper = blobhelper

    @storage_method
    @metricmethod
    def loadBlob(self, oid, serial):
        """Return the filename of the Blob data for this OID and serial.

        Returns a filename.

        Raises POSKeyError if the blobfile cannot be found.
        """
        cursor = self.load_connection.cursor
        return self.blobhelper.loadBlob(cursor, oid, serial)

    @storage_method
    @metricmethod
    def openCommittedBlobFile(self, oid, serial, blob=None):
        """
        Return a file for committed data for the given object id and serial

        If a blob is provided, then a BlobFile object is returned,
        otherwise, an ordinary file is returned. In either case, the
        file is opened for binary reading.

        This method is used to allow storages that cache blob data to
        make sure that data are available at least long enough for the
        file to be opened.
        """
        cursor = self.load_connection.cursor
        return self.blobhelper.openCommittedBlobFile(
            cursor, oid, serial, blob=blob)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Support for methods that require access to
an object's history. These only do useful things
when the database is history preserving.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from persistent.timestamp import TimeStamp

from ZODB.UndoLogCompatible import UndoLogCompatible
from ZODB.POSException import POSKeyError
from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64

from relstorage._compat import base64_encodebytes
from relstorage._compat import loads

from .util import phase_dependent
from .util import storage_method
from .util import stale_aware


class History(object):
    """
    Provides the implementation of ``history``
    defined in :class:`ZODB.interfaces.IStorage`.

    This is available for all database types.
    """

    __slots__ = (
        'adapter',
        'load_connection',
        '__dict__',
    )

    def __init__(self, adapter, load_connection):
        self.adapter = adapter
        self.load_connection = load_connection

    @stale_aware
    @storage_method
    def history(self, oid, version=None, size=1, filter=None):
        # pylint:disable=unused-argument,too-many-locals
        cursor = self.load_connection.cursor
        oid_int = bytes8_to_int64(oid)
        try:
            history = self.adapter.dbiter.iter_object_history(
                cursor, oid_int)
        except KeyError:
            raise POSKeyError(oid)

        res = []
        for entry in history:
            tid = int64_to_8bytes(entry.tid_int)
            if entry.extension:
                d = loads(entry.extension)
            else:
                d = {}
            d.update({
                "time": TimeStamp(tid).timeTime(),
                "user_name": entry.username or b'',
                "description": entry.description or b'',
                "tid": tid,
                "version": '',
                "size": entry.pickle_size,
                "rs_tid_int": entry.tid_int,
                "rs_oid_int": oid_int,
            })
            if filter is None or filter(d):
                res.append(d)
                if size is not None and len(res) >= size:
                    break
        return res


class UndoableHistory(History):
    """
    Provides the implementation of the unique methods
    defined in :class:`ZODB.interfaces.IStorageUndoable`.
    """

    __slots__ = ()

    @storage_method
    def undoInfo(self, *args, **kwargs):
        # UndoLogCompatible provides the
        # implementation of undoInfo using self.undoLog
        log = UndoLogCompatible()
        log.undoLog = self.undoLog
        return log.undoInfo(*args, **kwargs)

    @storage_method
    def supportsUndo(self):
        return self.adapter.keep_history

    supportsTransactionalUndo = supportsUndo

    @stale_aware
    @storage_method
    def undoLog(self, first=0, last=-20, filter=None):
        if last < 0:
            last = first - last

        # use a private connection to ensure the most current results
        with self.load_connection.isolated_connection() as cursor:
            tx_iter = self.adapter.dbiter.iter_transactions(cursor)
            i = 0
            res = []
            for tx in tx_iter:
                tid = int64_to_8bytes(tx.tid_int)
                # Note that user and desc are schizophrenic. The transaction
                # interface specifies that they are a Python str, *probably*
                # meaning bytes. But code in the wild and the ZODB test suite
                # sets them as native strings, meaning unicode on Py3. OTOH, the
                # test suite checks that this method *returns* them as bytes!
                # This is largely cleaned up with transaction 2.0/ZODB 5, where the storage
                # interface is defined in terms of bytes only.
                d = {
                    'id': base64_encodebytes(tid)[:-1],  # pylint:disable=deprecated-method
                    'time': TimeStamp(tid).timeTime(),
                    'user_name':  tx.username or b'',
                    'description': tx.description or b'',
                }
                if tx.extension:
                    d.update(loads(tx.extension))

                if filter is None or filter(d):
                    if i >= first:
                        res.append(d)
                    i += 1
                    if i >= last:
                        break
            return res

    @phase_dependent
    @storage_method
    def undo(self, tpc_phase, transaction_id, transaction):
        """
        Undo a transaction identified by transaction_id.

        transaction_id is the base 64 encoding of an 8 byte tid. Undo
        by writing new data that reverses the action taken by the
        transaction.
        """
        # This is called directly from code in DB.py on a new instance
        # (created either by new_instance() or a special
        # undo_instance()). That new instance is never asked to load
        # anything, or poll invalidations, so our storage cache is ineffective
        # (unless we had loaded persistent state files)
        #
        # TODO: Implement 'undo_instance' to make this clear.
        #
        # A regular Connection going through two-phase commit will
        # call tpc_begin(), do a bunch of store() from its commit(),
        # then tpc_vote(), tpc_finish().
        #
        # During undo, we get a tpc_begin(), then a bunch of undo() from
        # ZODB.DB.TransactionalUndo.commit(), then tpc_vote() and tpc_finish().
        tpc_phase.undo(transaction_id, transaction)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Locker implementations.
"""

from __future__ import absolute_import

from zope.interface import implementer


from ..interfaces import ILocker
from ..interfaces import UnableToAcquirePackUndoLockError
from ..locker import AbstractLocker


@implementer(ILocker)
class OracleLocker(AbstractLocker):

    def __init__(self, options, driver, inputsize_NUMBER, batcher_factory):
        super(OracleLocker, self).__init__(
            options=options,
            driver=driver,
            batcher_factory=batcher_factory
        )
        self.inputsize_NUMBER = inputsize_NUMBER
        timeout = ' WAIT ' + str(self.commit_lock_timeout)
        self._lock_current_clause = self._lock_current_clause + timeout
        # 'FOR SHARE' isn't supported so all locks are exclusive.
        self._lock_share_clause = self._lock_current_clause
        self._lock_share_clause_nowait = 'FOR UPDATE NOWAIT'

    def release_commit_lock(self, cursor):
        # no action needed
        pass

    def hold_pack_lock(self, cursor):
        """Try to acquire the pack lock.

        Raise an exception if packing or undo is already in progress.
        """
        stmt = """
        LOCK TABLE pack_lock IN EXCLUSIVE MODE NOWAIT
        """
        try:
            cursor.execute(stmt)
        except self.lock_exceptions:  # cx_Oracle.DatabaseError:
            raise UnableToAcquirePackUndoLockError('A pack or undo operation is in progress')

    def release_pack_lock(self, cursor):
        """Release the pack lock."""
        # No action needed



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
IOIDAllocator implementations.
"""

from __future__ import absolute_import

from zope.interface import implementer

from ..._compat import metricmethod
from ..interfaces import IOIDAllocator
from ..oidallocator import AbstractRangedOIDAllocator


@implementer(IOIDAllocator)
class PostgreSQLOIDAllocator(AbstractRangedOIDAllocator):

    __slots__ = ()

    def _set_min_oid_from_range(self, cursor, n):
        # This potentially wastes a value from the sequence to find
        # out that we're already past the max. But that's the only option:
        # curval() is local to this connection, and 'ALTER SEQUENCE ... MINVALUE n STARTS WITH n'
        # takes a lock and doesn't let anyone use nextval() until we commit
        # (which could take some time). (Connections are expensive in PostgreSQL,
        # so we don't want to do what Oracle does and execute ALTER in a new connection.)
        cursor.execute("""
        SELECT CASE WHEN %s > nextval('zoid_seq')
            THEN setval('zoid_seq', %s)
            ELSE 0
            END
        """, (n, n))

    @metricmethod
    def new_oids(self, cursor):
        """Return a sequence of new, unused OIDs."""
        stmt = "SELECT NEXTVAL('zoid_seq')"
        cursor.execute(stmt)
        n = cursor.fetchone()[0]
        return self._oid_range_around(n)

    def reset_oid(self, cursor):
        raise NotImplementedError



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostgreSQL adapter for RelStorage."""
from __future__ import absolute_import
from __future__ import print_function

import logging

from ..._compat import metricmethod
from ..connmanager import AbstractConnectionManager

logger = logging.getLogger(__name__)


class Psycopg2ConnectionManager(AbstractConnectionManager):

    def __init__(self, driver, dsn, options):
        self._dsn = dsn
        self.isolation_read_committed = driver.ISOLATION_LEVEL_READ_COMMITTED
        self.isolation_serializable = driver.ISOLATION_LEVEL_SERIALIZABLE
        self.isolation_repeatable_read = driver.ISOLATION_LEVEL_REPEATABLE_READ
        self.keep_history = options.keep_history
        self._db_connect_with_isolation = driver.connect_with_isolation
        super(Psycopg2ConnectionManager, self).__init__(options, driver)


    def _alter_dsn(self, replica):
        """Alter the DSN to use the specified replica.

        The replica parameter is a string specifying either host or host:port.
        """
        if ':' in replica:
            host, port = replica.split(':')
            dsn = '%s host=%s port=%s' % (self._dsn, host, port)
        else:
            dsn = '%s host=%s' % (self._dsn, replica)
        return dsn

    @metricmethod
    def open(self, isolation=None, deferrable=False, read_only=False,
             replica_selector=None, application_name=None, **kwargs):
        """Open a database connection and return (conn, cursor)."""
        # pylint:disable=arguments-differ
        if isolation is None:
            isolation = self.isolation_store

        if replica_selector is None:
            replica_selector = self.replica_selector

        if replica_selector is not None:
            replica = replica_selector.current()
            dsn = self._alter_dsn(replica)
        else:
            replica = None
            dsn = self._dsn

        while True:
            try:
                # psycopg2 seems to have a cache of Connection objects
                # so closing one and then opening again often gets the same
                # object back.
                conn = self._db_connect_with_isolation(
                    dsn,
                    isolation=isolation,
                    deferrable=deferrable,
                    read_only=read_only,
                    application_name=application_name
                )
                cursor = self.cursor_for_connection(conn)
                conn.replica = replica
                return conn, cursor
            except self.driver.use_replica_exceptions as e:
                if replica is not None:
                    next_replica = replica_selector.next()
                    if next_replica is not None:
                        logger.warning("Unable to connect to replica %s: %s, "
                                       "now trying %s", replica, e, next_replica)
                        replica = next_replica
                        dsn = self._alter_dsn(replica)
                        continue
                logger.warning("Unable to connect: %s", e)
                raise

    def _do_open_for_load(self):
        # In RelStorage 1, 2 and <= 3.0b2, we used SERIALIZABLE isolation,
        # while MySQL used REPEATABLE READ and Oracle used SERIALIZABLE (but
        # only because of an apparent issue with RAC).
        #
        # Although SERIALIZABLE is much cheaper on PostgreSQL than
        # most other databases, it has its issues. Most notably,
        # SERIALIZABLE isn't allowed on streaming replicas
        # (https://www.enterprisedb.com/blog/serializable-postgresql-11-and-beyond),
        # and prior to PostgreSQL 12 it disables parallel queries (not
        # that we expect many queries to be something that can benefit
        # from parallel workers.)
        #
        # The differences that SERIALIZABLE brings shouldn't be
        # relevant as we don't run the write transactions at that
        # level, and we never try to commit this transaction. So it's
        # mostly just overhead for tracking read anomalies that can
        # never happen. And the standby issue became a problem
        # (https://github.com/zodb/relstorage/issues/376) and we
        # dropped down to REPEATABLE READ.

        # Of course, there's a chance that if we could get the store
        # connections to work in SERIALIZABLE mode, we'd be able to
        # stop the explicit locking altogether. With judicious use of
        # savepoints, and proper re-raising of ConflictError, that
        # might be possible.

        # Using READ ONLY mode lets transactions (especially
        # SERIALIZABLE) elide some locks. If we were SERIALIZABLE,
        # we'd probably also want to enable deferrable transactions as
        # there's special support to make them cheaper (but they might
        # have to wait on other serializable transactions, but since
        # our only other serializable transactions would be READ ONLY
        # that shouldn't matter.)

        return self.open(
            self.isolation_repeatable_read,
            read_only=True,
            deferrable=False,
            replica_selector=self.ro_replica_selector,
            application_name='RS: Load'
        )
